# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional, Tuple, Union

import jax.numpy as jnp
from jax import custom_jvp, vmap

from apax.utils.jax_md_reduced.util import Array, f32, safe_mask

# Types


DisplacementFn = Callable[[Array, Array], Array]
MetricFn = Callable[[Array, Array], float]
DisplacementOrMetricFn = Union[DisplacementFn, MetricFn]

ShiftFn = Callable[[Array, Array], Array]

Space = Tuple[DisplacementFn, ShiftFn]
Box = Array

# Primitive Spatial Transforms


def inverse(box: Box) -> Box:
    """Compute the inverse of an affine transformation."""
    if jnp.isscalar(box) or box.size == 1:
        return 1 / box
    elif box.ndim == 1:
        return 1 / box
    elif box.ndim == 2:
        return jnp.linalg.inv(box)
    raise ValueError(f"Box must be either: a scalar, a vector, or a matrix. Found {box}.")


def _get_free_indices(n: int) -> str:
    return "".join([chr(ord("a") + i) for i in range(n)])


def raw_transform(box: Box, R: Array) -> Array:
    """Apply an affine transformation to positions.

    See `periodic_general` for a description of the semantics of `box`.

    Args:
      box: An affine transformation described in `periodic_general`.
      R: Array of positions. Should have  shape `(..., spatial_dimension)`.

    Returns:
      A transformed array positions of shape `(..., spatial_dimension)`.
    """
    if jnp.isscalar(box) or box.size == 1:
        return R * box
    elif box.ndim == 1:
        indices = _get_free_indices(R.ndim - 1) + "i"
        return jnp.einsum(f"i,{indices}->{indices}", box, R)
    elif box.ndim == 2:
        free_indices = _get_free_indices(R.ndim - 1)
        left_indices = free_indices + "j"
        right_indices = free_indices + "i"
        return jnp.einsum(f"ij,{left_indices}->{right_indices}", box, R)
    raise ValueError(f"Box must be either: a scalar, a vector, or a matrix. Found {box}.")


@custom_jvp
def transform(box: Box, R: Array) -> Array:
    """Apply an affine transformation to positions.

    See `periodic_general` for a description of the semantics of `box`.

    Args:
      box: An affine transformation described in `periodic_general`.
      R: Array of positions. Should have  shape `(..., spatial_dimension)`.

    Returns:
      A transformed array positions of shape `(..., spatial_dimension)`.
    """
    return raw_transform(box, R)


@transform.defjvp
def transform_jvp(primals, tangents):
    box, R = primals
    dbox, dR = tangents
    return (transform(box, R), dR + transform(dbox, R))


def pairwise_displacement(Ra: Array, Rb: Array) -> Array:
    """Compute a matrix of pairwise displacements given two sets of positions.

    Args:
      Ra: Vector of positions; `ndarray(shape=[spatial_dim])`.
      Rb: Vector of positions; `ndarray(shape=[spatial_dim])`.

    Returns:
      Matrix of displacements; `ndarray(shape=[spatial_dim])`.
    """
    if len(Ra.shape) != 1:
        msg = (
            "Can only compute displacements between vectors. To compute "
            "displacements between sets of vectors use vmap or TODO."
        )
        raise ValueError(msg)

    if Ra.shape != Rb.shape:
        msg = "Can only compute displacement between vectors of equal dimension."
        raise ValueError(msg)

    return Ra - Rb


def periodic_shift(side: Box, R: Array, dR: Array) -> Array:
    """Shifts positions, wrapping them back within a periodic hypercube."""
    return jnp.mod(R + dR, side)


def free() -> Space:
    """Free boundary conditions."""

    def displacement_fn(
        Ra: Array, Rb: Array, perturbation: Optional[Array] = None, **unused_kwargs
    ) -> Array:
        dR = pairwise_displacement(Ra, Rb)
        if perturbation is not None:
            dR = raw_transform(perturbation, dR)
        return dR

    def shift_fn(R: Array, dR: Array, **unused_kwargs) -> Array:
        return R + dR

    return displacement_fn, shift_fn


def periodic_displacement(side: Box, dR: Array) -> Array:
    """Wraps displacement vectors into a hypercube.

    Args:
      side: Specification of hypercube size. Either,
        (a) float if all sides have equal length.
        (b) ndarray(spatial_dim) if sides have different lengths.
      dR: Matrix of displacements; `ndarray(shape=[..., spatial_dim])`.
    Returns:
      Matrix of wrapped displacements; `ndarray(shape=[..., spatial_dim])`.
    """
    return jnp.mod(dR + side * f32(0.5), side) - f32(0.5) * side


def square_distance(dR: Array) -> Array:
    """Computes square distances.

    Args:
      dR: Matrix of displacements; `ndarray(shape=[..., spatial_dim])`.
    Returns:
      Matrix of squared distances; `ndarray(shape=[...])`.
    """
    return jnp.sum(dR**2, axis=-1)


def periodic_general(
    box: Box, fractional_coordinates: bool = True, wrapped: bool = True
) -> Space:
    r"""Periodic boundary conditions on a parallelepiped.

    This function defines a simulation on a parallelepiped, :math:`X`, formed by
    applying an affine transformation, :math:`T`, to the unit hypercube
    :math:`U = [0, 1]^d` along with periodic boundary conditions across all
    of the faces.

    Formally, the space is defined such that :math:`X = {Tu : u \in [0, 1]^d}`.

    The affine transformation, :math:`T`, can be specified in a number of different
    ways. For a parallelepiped that is: 1) a cube of side length :math:`L`, the affine
    transformation can simply be a scalar; 2) an orthorhombic unit cell can be
    specified by a vector `[Lx, Ly, Lz]` of lengths for each axis; 3) a general
    triclinic cell can be specified by an upper triangular matrix.

    There are a number of ways to parameterize a simulation on :math:`X`.
    `periodic_general` supports two parametrizations of :math:`X` that can be selected
    using the `fractional_coordinates` keyword argument.

      1) When `fractional_coordinates=True`, particle positions are stored in the
         unit cube, :math:`u\in U`. Here, the displacement function computes the
         displacement between :math:`x, y \in X` as :math:`d_X(x, y) = Td_U(u, v)` where
         :math:`d_U` is the displacement function on the unit cube, :math:`U`, :math:`x = Tu`, and
         :math:`v = Tv` with :math:`u, v \in U`. The derivative of the displacement function
         is defined so that derivatives live in :math:`X` (as opposed to being
         backpropagated to :math:`U`). The shift function, `shift_fn(R, dR)` is defined
         so that :math:`R` is expected to lie in :math:`U` while :math:`dR` should lie in :math:`X`. This
         combination enables code such as `shift_fn(R, force_fn(R))` to work as
         intended.

      2) When `fractional_coordinates=False`, particle positions are stored in
         the parallelepiped :math:`X`. Here, for :math:`x, y \in X`, the displacement function
         is defined as :math:`d_X(x, y) = Td_U(T^{-1}x, T^{-1}y)`. Since there is an
         extra multiplication by :math:`T^{-1}`, this parameterization is typically
         slower than `fractional_coordinates=False`. As in 1), the displacement
         function is defined to compute derivatives in :math:`X`. The shift function
         is defined so that :math:`R` and :math:`dR` should both lie in :math:`X`.

    Example:

    .. code-block:: python

       from jax import random
       side_length = 10.0
       disp_frac, shift_frac = periodic_general(side_length,
                                                 fractional_coordinates=True)
       disp_real, shift_real = periodic_general(side_length,
                                                 fractional_coordinates=False)

       # Instantiate random positions in both parameterizations.
       R_frac = random.uniform(random.PRNGKey(0), (4, 3))
       R_real = side_length * R_frac

       # Make some shift vectors.
       dR = random.normal(random.PRNGKey(0), (4, 3))

       disp_real(R_real[0], R_real[1]) == disp_frac(R_frac[0], R_frac[1])
       transform(side_length, shift_frac(R_frac, 1.0)) == shift_real(R_real, 1.0)

    It is often desirable to deform a simulation cell either: using a finite
    deformation during a simulation, or using an infinitesimal deformation while
    computing elastic constants. To do this using fractional coordinates, we can
    supply a new affine transformation as `displacement_fn(Ra, Rb, box=new_box)`.
    When using real coordinates, we can specify positions in a space :math:`X` defined
    by an affine transformation :math:`T` and compute displacements in a deformed space
    :math:`X'` defined by an affine transformation :math:`T'`. This is done by writing
    `displacement_fn(Ra, Rb, new_box=new_box)`.

    There are a few caveats when using `periodic_general`. `periodic_general`
    uses the minimum image convention, and so it will fail for potentials whose
    cutoff is longer than the half of the side-length of the box. It will also
    fail to find the correct image when the box is too deformed. We hope to add a
    more robust box for small simulations soon (TODO) along with better error
    checking. In the meantime caution is recommended.

    Args:
      box: A `(spatial_dim, spatial_dim)` affine transformation.
      fractional_coordinates: A boolean specifying whether positions are stored
        in the parallelepiped or the unit cube.
      wrapped: A boolean specifying whether or not particle positions are
        remapped back into the box after each step
    Returns:
      `(displacement_fn, shift_fn)` tuple.
    """
    inv_box = inverse(box)

    def displacement_fn(Ra, Rb, perturbation=None, **kwargs):
        _box, _inv_box = box, inv_box

        if "box" in kwargs:
            _box = kwargs["box"]

            if not fractional_coordinates:
                _inv_box = inverse(_box)

        if "new_box" in kwargs:
            _box = kwargs["new_box"]

        if not fractional_coordinates:
            Ra = transform(_inv_box, Ra)
            Rb = transform(_inv_box, Rb)

        dR = periodic_displacement(f32(1.0), pairwise_displacement(Ra, Rb))
        dR = transform(_box, dR)

        if perturbation is not None:
            dR = raw_transform(perturbation, dR)

        return dR

    def u(R, dR):
        if wrapped:
            return periodic_shift(f32(1.0), R, dR)
        return R + dR

    def shift_fn(R, dR, **kwargs):
        if not fractional_coordinates and not wrapped:
            return R + dR

        _box, _inv_box = box, inv_box
        if "box" in kwargs:
            _box = kwargs["box"]
            _inv_box = inverse(_box)

        if "new_box" in kwargs:
            _box = kwargs["new_box"]

        dR = transform(_inv_box, dR)
        if not fractional_coordinates:
            R = transform(_inv_box, R)

        R = u(R, dR)

        if not fractional_coordinates:
            R = transform(_box, R)
        return R

    return displacement_fn, shift_fn


def distance(dR: Array) -> Array:
    """Computes distances.

    Args:
      dR: Matrix of displacements; `ndarray(shape=[..., spatial_dim])`.
    Returns:
      Matrix of distances; `ndarray(shape=[...])`.
    """
    dr = square_distance(dR)
    return safe_mask(dr > 0, jnp.sqrt, dr)


def map_bond(metric_or_displacement: DisplacementOrMetricFn) -> DisplacementOrMetricFn:
    """Vectorizes a metric or displacement function over bonds."""
    return vmap(metric_or_displacement, (0, 0), 0)



import logging

import numpy as np

from apax.utils.data import load_data, split_atoms, split_idxs

log = logging.getLogger(__name__)


def load_data_files(data_config):
    """
    Load data files for training and validation.

    Parameters
    ----------
    data_config : object
        Data configuration object.

    Returns
    -------
    Tuple
        Tuple containing list of ase.Atoms objects for training and validation.
    """

    log.info("Running Input Pipeline")
    if data_config.data_path is not None:
        log.info(f"Reading data file {data_config.data_path}")
        atoms_list = load_data(data_config.data_path)

        train_idxs, val_idxs = split_idxs(
            atoms_list, data_config.n_train, data_config.n_valid
        )
        train_atoms_list, val_atoms_list = split_atoms(atoms_list, train_idxs, val_idxs)
        log.info(f"Found n_train: {len(train_atoms_list)}, n_val: {len(val_atoms_list)}")

        np.savez(
            data_config.model_version_path / "train_val_idxs",
            train_idxs=train_idxs,
            val_idxs=val_idxs,
        )

    elif data_config.train_data_path and data_config.val_data_path is not None:
        log.info(f"Reading training data file {data_config.train_data_path}")
        log.info(f"Reading validation data file {data_config.val_data_path}")
        train_atoms_list = load_data(data_config.train_data_path)
        val_atoms_list = load_data(data_config.val_data_path)
        log.info(f"Found n_train: {len(train_atoms_list)}, n_val: {len(val_atoms_list)}")
    else:
        raise ValueError("input data path/paths not defined")

    return train_atoms_list, val_atoms_list



import logging
import pathlib
import typing as t

import ase
import numpy as np
import pandas as pd
import yaml
import znh5md
import zntrack
from ase import Atoms

from apax.calibration import compute_calibration_factors
from apax.md import ASECalculator
from apax.md.function_transformations import GlobalCalibration, available_transformations
from apax.train.run import run as apax_run

from .utils import check_duplicate_keys

log = logging.getLogger(__name__)


class ApaxBase(zntrack.Node):
    def get_calculator(self, **kwargs):
        raise NotImplementedError


class Apax(ApaxBase):
    """Class for traing Apax models

    Parameters
    ----------
    config: str
        path to the apax config file
    data: list[ase.Atoms]
        the training data set
    validation_data: list[ase.Atoms]
        atoms object with the validation data set
    model: t.Optional[Apax]
        model to be used as a base model for transfer learning
    log_level: str
        verbosity of logging during training
    """

    data: list[ase.Atoms] = zntrack.deps()
    config: str = zntrack.params_path()
    validation_data: list[ase.Atoms] = zntrack.deps()
    model: t.Optional[ApaxBase] = zntrack.deps(None)
    nl_skin: float = zntrack.params(0.5)
    transformations: t.Optional[list[dict[str, dict]]] = zntrack.params(None)
    log_level: str = zntrack.params("info")

    model_directory: pathlib.Path = zntrack.outs_path(zntrack.nwd / "apax_model")

    train_data_file: pathlib.Path = zntrack.outs_path(zntrack.nwd / "train_atoms.h5")
    validation_data_file: pathlib.Path = zntrack.outs_path(zntrack.nwd / "val_atoms.h5")

    metrics: dict = zntrack.metrics()

    @property
    def parameter(self) -> dict:
        parameter = yaml.safe_load(self.state.fs.read_text(self.config))

        custom_parameters = {
            "directory": self.model_directory.as_posix(),
            "experiment": "",
            "train_data_path": self.train_data_file.as_posix(),
            "val_data_path": self.validation_data_file.as_posix(),
        }

        if self.model is not None:
            param_files = self.model.parameter["data"]["directory"]
            base_path = {"base_model_checkpoint": param_files}
            try:
                parameter["checkpoints"].update(base_path)
            except KeyError:
                parameter["checkpoints"] = base_path

        check_duplicate_keys(custom_parameters, parameter["data"], log)
        parameter["data"].update(custom_parameters)
        return parameter

    def train_model(self):
        """Train the model using `apax.train.run`"""
        apax_run(self.parameter, log_level=self.log_level)

    def get_metrics(self):
        """In addition to the plots write a model metric"""
        metrics_df = pd.read_csv(self.model_directory / "log.csv")
        best_epoch = np.argmin(metrics_df["val_loss"])
        self.metrics = metrics_df.iloc[best_epoch].to_dict()

    def run(self):
        """Primary method to run which executes all steps of the model training"""

        if not self.state.restarted:
            train_db = znh5md.IO(self.train_data_file.as_posix())
            train_db.extend(self.data)

            val_db = znh5md.IO(self.validation_data_file.as_posix())
            val_db.extend(self.validation_data)

        csv_path = self.model_directory / "log.csv"
        if self.state.restarted and csv_path.is_file():
            metrics_df = pd.read_csv(self.model_directory / "log.csv")

            if metrics_df["epoch"].iloc[-1] >= self.parameter["n_epochs"] - 1:
                return

        self.train_model()
        self.get_metrics()

    def get_calculator(self, **kwargs):
        """Get an apax ase calculator"""
        transformations = []
        if self.transformations:
            for transform, params in self.transformations.items():
                transformations.append(available_transformations[transform](**params))

        with self.state.use_tmp_path():
            calc = ASECalculator(
                model_dir=self.model_directory,
                dr_threshold=self.nl_skin,
                transformations=transformations,
            )
            return calc


class ApaxEnsemble(ApaxBase):
    """Parallel apax model ensemble in ASE.

    Parameters
    ----------
    models: list
        List of `ApaxModel` nodes to ensemble.
    nl_skin: float
        Neighborlist skin.
    transformations: dict
        Key-parameter dict with function transformations applied
        to the model function within the ASE calculator.
        See the apax documentation for available methods.
    """

    models: list[Apax] = zntrack.deps()
    nl_skin: float = zntrack.params(0.5)
    transformations: t.Optional[list[dict[str, dict]]] = zntrack.params(None)

    def run(self) -> None:
        pass

    def get_calculator(self, **kwargs) -> ase.calculators.calculator.Calculator:
        """Property to return a model specific ase calculator object.

        Returns
        -------
        calc:
            ase calculator object
        """
        param_files = [m.parameter["data"]["directory"] for m in self.models]

        transformations = []
        if self.transformations:
            for transform, params in self.transformations.items():
                transformations.append(available_transformations[transform](**params))

        with self.state.use_tmp_path():
            calc = ASECalculator(
                param_files,
                dr_threshold=self.nl_skin,
                transformations=transformations,
            )
            return calc


class ApaxImport(zntrack.Node):
    """Parallel apax model ensemble in ASE.

    Parameters
    ----------
    models: list
        List of `ApaxModel` nodes to ensemble.
    nl_skin: float
        Neighborlist skin.
    transformations: dict
        Key-parameter dict with function transformations applied
        to the model function within the ASE calculator.
        See the apax documentation for available methods.
    """

    config: str = zntrack.params_path()
    nl_skin: float = zntrack.params(0.5)
    transformations: t.Optional[list[dict[str, dict]]] = zntrack.params(None)

    @property
    def parameter(self) -> dict:
        return yaml.safe_load(self.state.fs.read_text(self.config))

    def get_calculator(self, **kwargs) -> ase.calculators.calculator.Calculator:
        """Property to return a model specific ase calculator object.

        Returns
        -------
        calc:
            ase calculator object
        """

        directory = self.parameter["data"]["directory"]
        exp = self.parameter["data"]["experiment"]
        model_dir = directory + "/" + exp

        transformations = []
        if self.transformations:
            for transform, params in self.transformations.items():
                transformations.append(available_transformations[transform](**params))

        with self.state.use_tmp_path():
            calc = ASECalculator(
                model_dir,
                dr=self.nl_skin,
                transformations=transformations,
            )
            return calc

    def run(self) -> None:
        pass


class ApaxCalibrate(ApaxBase):
    """Globally calibrate the energy and force uncertainties of an Apax model.

    Parameters
    ----------
    model: Apax
        Model to calibrate
    validation_data: list[ase.Atoms]
        Calibration atoms
    batch_size: int, default = 32
        Processing batch size. Choose the largest allowed by your VRAM.
    criterion: str, default = "ma_cal
        Calibration criterion. See uncertainty_toolbox for more details.
    shared_factor: bool, default = False
        Whether or not to calibrate energies and forces separately.
    optimizer_bounds: Tuple[float, float], default = (1e-2, 1e2)
        Search value bounds.
    transformations: dict
        Key-parameter dict with function transformations applied
        to the model function within the ASE calculator.
        See the apax documentation for available methods.
    """

    model: ApaxBase = zntrack.deps()
    validation_data: list[Atoms] = zntrack.deps()
    batch_size: int = zntrack.params(32)
    criterion: str = zntrack.params("ma_cal")
    shared_factor: bool = zntrack.params(False)
    optimizer_bounds: t.Tuple[float, float] = zntrack.params((1e-2, 1e2))

    transformations: t.Optional[list[dict[str, dict]]] = zntrack.params(None)

    nl_skin: float = zntrack.params(0.5)

    metrics: dict = zntrack.metrics()

    def run(self):
        """Primary method to run which executes all steps of the model training"""

        calc = self.model.get_calculator()
        self.e_factor, self.f_factor = compute_calibration_factors(
            calc,
            self.validation_data,
            batch_size=self.batch_size,
            criterion=self.criterion,
            shared_factor=self.shared_factor,
            optimizer_bounds=self.optimizer_bounds,
        )

        self.metrics = {
            "e_factor": self.e_factor,
            "f_factor": self.f_factor,
        }

    def get_calculator(self, **kwargs) -> ase.calculators.calculator.Calculator:
        """Property to return a model specific ase calculator object.

        Returns
        -------
        calc:
            ase calculator object
        """

        e_factor = self.metrics["e_factor"]
        f_factor = self.metrics["f_factor"]

        config_file = self.model.parameter["data"]["directory"]

        calibration = GlobalCalibration(
            energy_factor=e_factor,
            forces_factor=f_factor,
        )
        transformations = [calibration]
        if self.transformations:
            for transform, params in self.transformations.items():
                transformations.append(available_transformations[transform](**params))

        with self.state.use_tmp_path():
            calc = ASECalculator(
                config_file,
                dr=self.nl_skin,
                transformations=transformations,
            )
            return calc



from .md import ApaxJaxMD
from .model import Apax, ApaxCalibrate, ApaxEnsemble, ApaxImport
from .utils import AddData

__all__ = ["Apax", "ApaxEnsemble", "ApaxJaxMD", "ApaxImport", "ApaxCalibrate", "AddData"]

try:
    from .analysis import ApaxBatchPrediction  # noqa: F401
    from .selection import BatchKernelSelection  # noqa: F401

    __all__.append("ApaxBatchPrediction")
    __all__.append("BatchKernelSelection")
except ImportError:
    pass



import logging
from pathlib import Path

import h5py
import numpy as np
import znh5md
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator

from apax.md.sim_utils import System
from apax.utils.helpers import APAX_PROPERTIES
from apax.utils.jax_md_reduced import space

log = logging.getLogger(__name__)


class TrajHandler:
    def __init__(
        self,
        system: System,
        sampling_rate: int,
        buffer_size: int,
        traj_path: Path,
        time_step: float = 0.5,
        properties: list[str] = APAX_PROPERTIES,
    ) -> None:
        self.atomic_numbers = system.atomic_numbers
        self.box = system.box
        self.fractional = np.any(self.box > 1e-6)
        self.sampling_rate = sampling_rate
        self.traj_path = traj_path
        self.time_step = time_step
        self.properties = properties

    def step(self, state_and_energy, transform=None):
        pass

    def write(self, x=None, transform=None):
        pass

    def close(self):
        pass

    def reset_buffer(self):
        pass

    def atoms_from_state(self, state, predictions, nbr_kwargs):
        if "box" in nbr_kwargs.keys():
            box = nbr_kwargs["box"]
        else:
            box = self.box

        if self.fractional:
            positions = space.transform(box, state.position)
        else:
            positions = state.position

        positions = np.asarray(positions)
        momenta = np.asarray(state.momentum)
        forces = np.asarray(state.force)

        atoms = Atoms(self.atomic_numbers, positions, momenta=momenta, cell=box)
        atoms.cell = atoms.cell.T
        atoms.pbc = np.diag(atoms.cell.array) > 1e-6
        predictions = {k: np.array(v) for k, v in predictions.items()}
        predictions["energy"] = predictions["energy"].item()
        predictions = {k: v for k, v in predictions.items() if k in self.properties}
        atoms.calc = SinglePointCalculator(atoms, **predictions)
        return atoms


class H5TrajHandler(TrajHandler):
    def __init__(
        self,
        system: System,
        sampling_rate: int,
        buffer_size: int,
        traj_path: Path,
        time_step: float,
        properties: list[str],
        h5md_options: dict,
    ) -> None:
        self.atomic_numbers = system.atomic_numbers
        self.box = system.box
        self.fractional = np.any(self.box > 1e-6)
        self.sampling_rate = sampling_rate
        self.traj_path = traj_path
        self.time_step = time_step
        self.properties = properties
        if h5md_options is None:
            h5md_options = {}
        self.db = znh5md.IO(
            self.traj_path,
            timestep=self.time_step,
            save_units=False,
            variable_shape=False,
            **h5md_options,
        )

        self.step_counter = 0
        self.buffer = []
        self.buffer_size = buffer_size

    def reset_buffer(self):
        self.buffer = []

    def step(self, state, transform=None):
        state, predictions, nbr_kwargs = state

        if self.step_counter % self.sampling_rate == 0:
            new_atoms = self.atoms_from_state(state, predictions, nbr_kwargs)
            self.buffer.append(new_atoms)
        self.step_counter += 1

        if len(self.buffer) >= self.buffer_size:
            self.write()

    def write(self, x=None, transform=None):
        if len(self.buffer) > 0:
            self.db.extend(self.buffer)
            self.reset_buffer()


class DSTruncator:
    def __init__(self, length):
        self.length = length
        self.node_names = []

    def __call__(self, name, node):
        if isinstance(node, h5py.Dataset):
            if len(node.shape) > 1 or name.endswith("energy/value"):
                self.node_names.append(name)

    def truncate(self, ds):
        for name in self.node_names:
            shape = tuple([None] + list(ds[name].shape[1:]))
            truncated_data = ds[name][: self.length]
            del ds[name]
            ds.create_dataset(name, maxshape=shape, data=truncated_data, chunks=True)


def truncate_trajectory_to_checkpoint(traj_path, length):
    truncator = DSTruncator(length=length)
    with h5py.File(traj_path, "r+") as ds:
        ds.visititems(truncator)
        truncator.truncate(ds)



import os
import pathlib
import uuid

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import yaml
import znh5md
from ase import Atoms
from ase.io import read, write
from flax.training import checkpoints

from apax.config import Config, MDConfig
from apax.md import run_md
from apax.md.ase_calc import ASECalculator
from apax.utils import jax_md_reduced
from tests.conftest import load_config_and_run_training

TEST_PATH = pathlib.Path(__file__).parent.resolve()


def test_run_md(get_tmp_path):
    model_confg_path = TEST_PATH / "config.yaml"
    md_confg_path = TEST_PATH / "md_config.yaml"

    with open(model_confg_path.as_posix(), "r") as stream:
        model_config_dict = yaml.safe_load(stream)
    with open(md_confg_path.as_posix(), "r") as stream:
        md_config_dict = yaml.safe_load(stream)

    model_config_dict["data"]["directory"] = get_tmp_path.as_posix()
    md_config_dict["sim_dir"] = get_tmp_path.as_posix()
    md_config_dict["initial_structure"] = get_tmp_path.as_posix() + "/atoms.extxyz"

    model_config = Config.model_validate(model_config_dict)
    os.makedirs(model_config.data.model_version_path)
    model_config.dump_config(model_config.data.model_version_path)
    md_config = MDConfig.model_validate(md_config_dict)

    positions = jnp.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ],
        dtype=jnp.float64,
    )
    atomic_numbers = np.array([1, 2, 2])
    box = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    offsets = jnp.full([3, 3], 0)
    atoms = Atoms(atomic_numbers, positions, cell=box)
    write(md_config.initial_structure, atoms)

    n_species = 119  # int(np.max(atomic_numbers) + 1)

    displacement_fn, _ = jax_md_reduced.space.free()

    neighbor_fn = jax_md_reduced.partition.neighbor_list(
        displacement_or_metric=displacement_fn,
        box=box,
        r_cutoff=model_config.model.basis.r_max,
        format=jax_md_reduced.partition.Sparse,
        fractional_coordinates=False,
    )
    neighbors = neighbor_fn.allocate(positions)

    Builder = model_config.model.get_builder()
    builder = Builder(model_config.model.model_dump(), n_species=n_species)
    model = builder.build_energy_derivative_model(
        apply_mask=False, inference_disp_fn=displacement_fn
    )
    rng_key = jax.random.PRNGKey(model_config.seed)
    params = model.init(
        rng_key,
        positions,
        atomic_numbers,
        neighbors.idx,
        box,
        offsets,
    )

    ckpt = {"model": {"params": params}, "epoch": 0}
    checkpoints.save_checkpoint(
        ckpt_dir=model_config.data.best_model_path,
        target=ckpt,
        step=0,
        overwrite=True,
    )

    run_md(model_config_dict, md_config_dict)

    traj = znh5md.IO(md_config.sim_dir + "/" + md_config.traj_name)[:]
    assert len(traj) == 5  # num steps

    # checkpoint loading
    md_config_dict["duration"] = 0.5

    run_md(model_config_dict, md_config_dict)

    traj = znh5md.IO(md_config.sim_dir + "/" + md_config.traj_name)[:]
    assert len(traj) == 10  # num steps


def test_ase_calc(get_tmp_path):
    model_confg_path = TEST_PATH / "config.yaml"
    initial_structure_path = get_tmp_path / "atoms.extxyz"

    with open(model_confg_path.as_posix(), "r") as stream:
        model_config_dict = yaml.safe_load(stream)

    model_config_dict["data"]["directory"] = get_tmp_path.as_posix()

    model_config = Config.model_validate(model_config_dict)
    os.makedirs(model_config.data.model_version_path, exist_ok=True)
    model_config.dump_config(model_config.data.model_version_path)

    cell_size = 10.0
    positions = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    atomic_numbers = np.array([1, 1, 8])
    box = np.diag([cell_size] * 3)
    offsets = jnp.full([3, 3], 0)
    atoms = Atoms(atomic_numbers, positions, cell=box)
    write(initial_structure_path.as_posix(), atoms)

    displacement_fn, _ = jax_md_reduced.space.periodic_general(
        cell_size, fractional_coordinates=False
    )

    neighbor_fn = jax_md_reduced.partition.neighbor_list(
        displacement_or_metric=displacement_fn,
        box=box,
        r_cutoff=model_config.model.basis.r_max,
        format=jax_md_reduced.partition.Sparse,
        fractional_coordinates=False,
    )
    neighbors = neighbor_fn.allocate(positions)

    Builder = model_config.model.get_builder()
    builder = Builder(model_config.model.model_dump())
    model = builder.build_energy_derivative_model(inference_disp_fn=displacement_fn)
    rng_key = jax.random.PRNGKey(model_config.seed)
    params = model.init(
        rng_key,
        jnp.asarray(positions, dtype=jnp.float32),
        jnp.asarray(atomic_numbers),
        neighbors.idx,
        box,
        offsets=offsets,
    )
    ckpt = {"model": {"params": params}, "epoch": 0}

    checkpoints.save_checkpoint(
        ckpt_dir=model_config.data.best_model_path,
        target=ckpt,
        step=0,
        overwrite=True,
    )

    atoms = read(initial_structure_path.as_posix())
    calc = ASECalculator(
        [model_config.data.model_version_path, model_config.data.model_version_path]
    )

    atoms.calc = calc
    E = atoms.get_potential_energy()
    F = atoms.get_forces()
    S = atoms.get_stress()

    assert E != 0
    assert F.shape == (3, 3)
    assert S.shape == (6,)

    assert "energy_uncertainty" in atoms.calc.results.keys()
    assert "forces_uncertainty" in atoms.calc.results.keys()
    assert "stress_uncertainty" in atoms.calc.results.keys()


@pytest.mark.parametrize("num_data", (30,))
def test_jaxmd_schedule_and_thresold(get_tmp_path, example_dataset):
    model_confg_path = TEST_PATH / "config.yaml"
    working_dir = get_tmp_path / str(uuid.uuid4())
    data_path = get_tmp_path / "ds.extxyz"

    write(data_path, example_dataset)

    data_config_mods = {
        "data": {
            "directory": working_dir.as_posix(),
            "experiment": "model",
            "data_path": data_path.as_posix(),
        },
    }
    model_config_dict = load_config_and_run_training(model_confg_path, data_config_mods)

    md_confg_path = TEST_PATH / "md_config_threshold.yaml"

    with open(md_confg_path.as_posix(), "r") as stream:
        md_config_dict = yaml.safe_load(stream)
    md_config_dict["sim_dir"] = get_tmp_path.as_posix()
    md_config_dict["initial_structure"] = get_tmp_path.as_posix() + "/ds.extxyz"
    md_config = MDConfig.model_validate(md_config_dict)

    model_config = Config.model_validate(model_config_dict)

    run_md(model_config, md_config)

    traj = znh5md.IO(md_config.sim_dir + "/" + md_config.traj_name)[:]
    assert len(traj) < 1000  # num steps

    results_keys = list(traj[0].calc.results.keys())

    assert "energy" in results_keys
    assert "forces" in results_keys
    assert "energy_uncertainty" in results_keys
    assert "forces_ensemble" in results_keys

    assert "energy_ensemble" not in results_keys
    assert "forces_uncertainty" not in results_keys



import logging
import os
import sys
from typing import List, Union

import jax

from apax.config import Config, LossConfig, parse_config
from apax.data.initialization import load_data_files
from apax.data.input_pipeline import dataset_dict
from apax.data.statistics import compute_scale_shift_parameters
from apax.optimizer import get_opt
from apax.train.callbacks import initialize_callbacks
from apax.train.checkpoints import create_params, create_train_state
from apax.train.loss import Loss, LossCollection
from apax.train.metrics import initialize_metrics
from apax.train.parameters import EMAParameters
from apax.train.trainer import fit
from apax.transfer_learning import transfer_parameters
from apax.utils.random import seed_py_np_tf

log = logging.getLogger(__name__)


def setup_logging(log_file, log_level):
    """
    Setup logging configuration.

    Parameters
    ----------
    log_file : str
        Path to the log file.
    log_level : str
        Logging level. Options: {'debug', 'info', 'warning', 'error', 'critical'}.
    """

    log_levels = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }

    while len(logging.root.handlers) > 0:
        logging.root.removeHandler(logging.root.handlers[-1])

    logging.getLogger("absl").setLevel(logging.WARNING)

    logging.basicConfig(
        level=log_levels[log_level],
        format="%(levelname)s | %(asctime)s | %(message)s",
        datefmt="%H:%M:%S",
        handlers=[logging.FileHandler(log_file), logging.StreamHandler(sys.stderr)],
    )


def initialize_loss_fn(loss_config_list: List[LossConfig]) -> LossCollection:
    """
    Initialize loss functions based on configuration.

    Parameters
    ----------
    loss_config_list : List[LossConfig]
        List of loss configurations.

    Returns
    -------
    LossCollection
        Collection of initialized loss functions.
    """

    log.info("Initializing Loss Function")
    loss_funcs = []
    for loss in loss_config_list:
        loss_funcs.append(Loss(**loss.model_dump()))
    return LossCollection(loss_funcs)


def compute_property_shapes(config: Config):
    property_configs = [p.model_dump() for p in config.model.property_heads]

    additional_properties = []

    if len(property_configs) == 0:
        return additional_properties

    loss_names = [loss.name for loss in config.loss]
    for pconf in property_configs:
        name = pconf["name"]
        if name not in loss_names:
            continue
        shape = []
        if pconf["aggregation"] == "none":
            shape.append("natoms")

        feature_shapes = {"l0": [1], "l1": [3], "symmetric_traceless_l2": [3, 3]}

        shape.extend(feature_shapes[pconf["mode"]])

        additional_properties.append((name, shape))

    return additional_properties


def initialize_datasets(config: Config):
    """
    Initialize training and validation datasets based on the provided configuration.

    Parameters
    ----------
    config : Config
        Configuration object all parameters.

    Returns
    -------
    train_ds : Dataset
        Training dataset.
    val_ds : Dataset
        Validation dataset.
    ds_stats : Dict[str, Tuple[float, float]]
        Dictionary containing scale and shift parameters for normalization.
    """

    train_raw_ds, val_raw_ds = load_data_files(config.data)

    Dataset = dataset_dict[config.data.dataset.processing]

    dataset_kwargs = dict(config.data.dataset)
    processing = dataset_kwargs.pop("processing")

    if processing == "cached":
        dataset_kwargs["cache_path"] = config.data.model_version_path

    additional_properties = compute_property_shapes(config)

    train_ds = Dataset(
        train_raw_ds,
        config.model.basis.r_max,
        config.data.batch_size,
        config.n_epochs,
        config.n_jitted_steps,
        pos_unit=config.data.pos_unit,
        energy_unit=config.data.energy_unit,
        pre_shuffle=True,
        additional_properties=additional_properties,
        **dataset_kwargs,
    )
    val_ds = Dataset(
        val_raw_ds,
        config.model.basis.r_max,
        config.data.valid_batch_size,
        config.n_epochs,
        pos_unit=config.data.pos_unit,
        energy_unit=config.data.energy_unit,
        additional_properties=additional_properties,
        **dataset_kwargs,
    )
    ds_stats = compute_scale_shift_parameters(
        train_ds.inputs,
        train_ds.labels,
        config.data.shift_method,
        config.data.scale_method,
        config.data.shift_options,
        config.data.scale_options,
    )
    return train_ds, val_ds, ds_stats


def run(user_config: Union[str, os.PathLike, dict], log_level="error"):
    """
    Starts the training of a model with parameters provided by a the config.

    Parameters
    ----------
    user_config : str | os.PathLike | dict
        training config full example can be find :ref:`here <train_config>`:

    """
    config = parse_config(user_config)

    seed_py_np_tf(config.seed)
    rng_key = jax.random.PRNGKey(config.seed)

    config.data.model_version_path.mkdir(parents=True, exist_ok=True)
    setup_logging(config.data.model_version_path / "train.log", log_level)
    config.dump_config(config.data.model_version_path)
    log.info(f"Running on {jax.devices()}")

    callbacks = initialize_callbacks(config, config.data.model_version_path)
    loss_fn = initialize_loss_fn(config.loss)
    Metrics = initialize_metrics(config.metrics)

    train_ds, val_ds, ds_stats = initialize_datasets(config)

    sample_input, init_box = train_ds.init_input()
    Builder = config.model.get_builder()
    builder = Builder(config.model.model_dump())
    model = builder.build_energy_derivative_model(
        scale=ds_stats.elemental_scale,
        shift=ds_stats.elemental_shift,
        apply_mask=True,
        init_box=init_box,
    )
    batched_model = jax.vmap(model.apply, in_axes=(None, 0, 0, 0, 0, 0))

    if config.model.ensemble and config.model.ensemble.kind == "full":
        n_full_models = config.model.ensemble.n_members
    else:
        n_full_models = 1
    params, rng_key = create_params(model, rng_key, sample_input, n_full_models)

    # TODO rework optimizer initialization and lr keywords
    steps_per_epoch = train_ds.steps_per_epoch()
    tx = get_opt(
        params,
        config.n_epochs,
        steps_per_epoch,
        **config.optimizer.model_dump(),
    )

    state = create_train_state(batched_model, params, tx)

    base_checkpoint = config.checkpoints.base_model_checkpoint
    do_transfer_learning = base_checkpoint is not None
    if do_transfer_learning:
        state = transfer_parameters(state, config.checkpoints)

    if config.weight_average:
        ema_handler = EMAParameters(
            config.weight_average.ema_start, config.weight_average.alpha
        )
    else:
        ema_handler = None

    fit(
        state,
        train_ds,
        loss_fn,
        Metrics,
        callbacks,
        config.n_epochs,
        ckpt_dir=config.data.model_version_path,
        ckpt_interval=config.checkpoints.ckpt_interval,
        val_ds=val_ds,
        patience=config.patience,
        disable_pbar=config.progress_bar.disable_epoch_pbar,
        disable_batch_pbar=config.progress_bar.disable_batch_pbar,
        is_ensemble=n_full_models > 1,
        data_parallel=config.data_parallel,
        ema_handler=ema_handler,
    )
    log.info("Finished training")



from typing import Optional

import numpy as np

from apax.bal.kernel import KernelMatrix


def max_dist_selection(matrix: KernelMatrix, batch_size: Optional[int] = None):
    """
    Iteratively selects samples from the pool which are
    most distant from all previously selected samples.
    \\argmax_{S \\in \\mathbb{X}_{rem}} \\min_{S' \\in \\mathbb{X}_{sel} } d(S, S')

    https://arxiv.org/pdf/2203.09410.pdf
    https://doi.org/10.1039/D2DD00034B

    Attributes
    ----------
    matrix: KernelMatrix
        Kernel used to compare structures.
    batch_size: int
        Number of new data points to be selected.
    """
    n_train = matrix.n_train

    min_squared_distances = matrix.diagonal
    min_squared_distances[:n_train] = -np.inf

    n_pool = len(min_squared_distances[n_train:])
    end = n_pool

    if batch_size:
        end = batch_size

    # Use max norm for first point
    new_idx = np.argmax(min_squared_distances)
    selected_idxs = list(range(n_train)) + [new_idx]
    distances = [np.max(min_squared_distances)]

    for _ in range(1, end):
        squared_distances = matrix.score(new_idx)

        squared_distances[selected_idxs] = -np.inf
        min_squared_distances = np.minimum(min_squared_distances, squared_distances)

        new_idx = np.argmax(min_squared_distances)
        max_dist = np.max(min_squared_distances)
        selected_idxs.append(new_idx)
        distances.append(max_dist)

    # shift by number of train datapoints
    selected_idxs = np.array(selected_idxs[n_train:]) - n_train
    distances = np.array(distances)
    return selected_idxs, distances



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../../"))

project = "Apax"
copyright = "2023, Moritz Schäfer, Nico Segreto, Johannes Kästner"
author = "Moritz Schäfer, Nico Segreto, Johannes Kästner"
release = "0.11.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx_copybutton",
    "nbsphinx",
    "nbsphinx_link",
    "sphinx.ext.viewcode",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"

html_theme_options = {}



from typing import Any

import e3x
import flax.linen as nn
import jax
import jax.numpy as jnp
from jaxtyping import Array
from myrto.so3krates.embedding import (
    ChemicalEmbedding,
    SphericalHarmonics,
)
from myrto.so3krates.so3krates import TransformerBlock
from myrto.utils.safe import masked

from apax.layers.descriptor.basis_functions import BesselBasis
from apax.utils.convert import str_to_dtype


def get_node_mask(Z):
    mask = (Z != 0).astype(jnp.int16)
    return mask


def get_neighbor_mask(idx):
    mask = ((idx[0] - idx[1]) != 0).astype(jnp.int16)
    return mask


class So3kratesRepresentation(nn.Module):
    basis_fn: nn.Module = BesselBasis()
    num_layers: int = 1
    max_degree: int = 3
    num_features: int = 128
    num_heads: int = 4
    use_layer_norm_1: bool = False
    use_layer_norm_2: bool = False
    use_layer_norm_final: bool = False
    activation: str = "silu"
    cutoff_fn: str = "cosine_cutoff"
    transform_input_features: bool = False
    dtype: Any = jnp.float32

    @nn.compact
    def __call__(
        self,
        dr_vec: Array,
        Z: Array,
        idx: Array,
    ):
        dtype = str_to_dtype(self.dtype)
        dr_vec = dr_vec.astype(dtype)

        R_ij = dr_vec
        i, j = idx[0], idx[1]
        Z_i = Z

        pair_mask = get_neighbor_mask(idx)
        node_mask = get_node_mask(Z)

        scalar_features = ChemicalEmbedding(num_features=self.num_features)(
            Z_i
        )  # -> [nodes, num_features]
        scalar_features *= node_mask[..., None]

        r_ij = e3x.ops.norm(R_ij, axis=-1)  # -> [pairs]

        cutoff_fn = getattr(e3x.nn.functions, self.cutoff_fn)
        cutoffs = cutoff_fn(r_ij, cutoff=self.basis_fn.r_max)  # -> [pairs]
        cutoffs *= pair_mask

        neighborhood_sizes = jax.ops.segment_sum(cutoffs, i, num_segments=Z_i.shape[0])[
            i
        ]  # -> [pairs]
        pair_scale = masked(
            lambda x: 1 / x, neighborhood_sizes[:, None], pair_mask
        ).reshape(-1)

        # radial_expansion = RadialEmbedding(self.num_radial_features, self.cutoff)(r_ij)
        radial_expansion = self.basis_fn(r_ij)
        radial_expansion *= cutoffs[..., None]  # -> [pairs, num_radial_features]

        spherical_expansion = SphericalHarmonics(self.max_degree)(
            R_ij
        )  # -> [pairs, (max_degree+1)**2]
        spherical_expansion *= pair_scale[..., None]

        spherical_features = jax.ops.segment_sum(
            spherical_expansion,
            i,
            num_segments=Z_i.shape[0],
        )  # -> [nodes, (max_degree+1)**2]
        spherical_features *= node_mask[..., None]

        for _ in range(self.num_layers):
            scalar_features, spherical_features = TransformerBlock(
                self.num_heads,
                self.use_layer_norm_1,
                self.use_layer_norm_2,
                activation=self.activation,
                transform_input_features=self.transform_input_features,
            )(
                scalar_features,
                spherical_features,
                radial_expansion,
                spherical_expansion,
                i,
                j,
                pair_mask,
                node_mask,
                cutoffs,
                pair_scale,
            )

        if self.use_layer_norm_final:
            scalar_features = masked(nn.LayerNorm(), scalar_features, node_mask)

        return scalar_features



import logging
from pathlib import Path
from typing import List, Tuple, Union

import jax
import jax.numpy as jnp
from flax.core.frozen_dict import FrozenDict, freeze, unfreeze
from flax.training import checkpoints, train_state
from flax.traverse_util import flatten_dict, unflatten_dict

from apax.config.common import parse_config
from apax.config.train_config import Config

log = logging.getLogger(__name__)


def check_for_ensemble(params: FrozenDict) -> int:
    """Checks if a set of parameters belongs to an ensemble model.
    This is the case if all parameters share the same first dimension (parameter batch)
    """
    flat_params = flatten_dict(params)
    shapes = [v.shape[0] for v in flat_params.values()]
    is_ensemble = len(set(shapes)) == 1

    if is_ensemble:
        return shapes[0]
    else:
        return 1


def create_train_state(model, params: FrozenDict, tx):
    n_models = check_for_ensemble(params)

    def create_single_train_state(params):
        state = train_state.TrainState.create(
            apply_fn=model,
            params=params,
            tx=tx,
        )
        return state

    if n_models > 1:
        train_state_fn = jax.vmap(create_single_train_state, axis_name="ensemble")
    else:
        train_state_fn = create_single_train_state

    return train_state_fn(params)


def create_params(model, rng_key, sample_input: tuple, n_models: int):
    keys = jax.random.split(rng_key, num=n_models + 1)
    rng_key, model_rng = keys[0], keys[1:]

    log.info(f"Initializing {n_models} model(s)")

    if n_models == 1:
        params = model.init(model_rng[0], *sample_input)
    elif n_models > 1:
        num_args = len(sample_input)
        # vmap only over parameters, not over any data from the input
        in_axes = (0, *[None] * num_args)
        params = jax.vmap(model.init, in_axes=in_axes)(model_rng, *sample_input)
    else:
        raise ValueError(f"n_models should be a positive integer, found {n_models}")

    params = freeze(params)

    return params, rng_key


def load_state(state, ckpt_dir):
    ckpt_dir = Path(ckpt_dir)
    start_epoch = 0
    target = {"model": state, "epoch": 0}
    checkpoints_exist = ckpt_dir.is_dir()
    if checkpoints_exist:
        log.info("Loading checkpoint")
        raw_restored = checkpoints.restore_checkpoint(
            ckpt_dir.resolve(), target=target, step=None
        )
        state = raw_restored["model"]
        start_epoch = raw_restored["epoch"] + 1
        log.info("Successfully restored checkpoint from epoch %d", raw_restored["epoch"])

    return state, start_epoch


class CheckpointManager:
    def __init__(self) -> None:
        self.async_manager = checkpoints.AsyncManager()

    def save_checkpoint(self, ckpt, epoch: int, path: Path) -> None:
        checkpoints.save_checkpoint(
            ckpt_dir=path.resolve(),
            target=ckpt,
            step=epoch,
            overwrite=True,
            keep=2,
            async_manager=self.async_manager,
        )


def stack_parameters(param_list: List[FrozenDict]) -> FrozenDict:
    """Combine a list of parameter sets into a stacked version.
    Used for model ensembles.
    """
    flat_param_list = []
    for params in param_list:
        params = unfreeze(params)
        flat_params = flatten_dict(params)
        flat_param_list.append(flat_params)

    stacked_flat_params = flat_params
    for p in flat_param_list[0].keys():
        stacked_flat_params[p] = jnp.stack(
            [flat_param[p] for flat_param in flat_param_list]
        )

    stacked_params = unflatten_dict(stacked_flat_params)
    stack_params = freeze(stacked_params)
    return stack_params


def load_params(model_version_path: Path, best=True) -> FrozenDict:
    model_version_path = Path(model_version_path)
    if best:
        model_version_path = model_version_path / "best"
    log.info(f"loading checkpoint from {model_version_path}")
    try:
        # keep try except block for zntrack load from rev
        raw_restored = checkpoints.restore_checkpoint(
            model_version_path, target=None, step=None
        )
    except FileNotFoundError:
        print(f"No checkpoint found at {model_version_path}")
    if raw_restored is None:
        raise FileNotFoundError(f"No checkpoint found at {model_version_path}")
    params = jax.tree.map(jnp.asarray, raw_restored["model"]["params"])

    return params


def restore_single_parameters(model_dir: Path) -> Tuple[Config, FrozenDict]:
    """Load the config and parameters of a single model"""
    model_dir = Path(model_dir)
    model_config = parse_config(model_dir / "config.yaml")

    if model_config.data.experiment == "":
        model_config.data.directory = model_dir.resolve().as_posix()
    else:
        model_config.data.directory = model_dir.parent.resolve().as_posix()

    ckpt_dir = model_config.data.model_version_path
    return model_config, load_params(ckpt_dir)


def restore_parameters(model_dir: Union[Path, List[Path]]) -> Tuple[Config, FrozenDict]:
    """Restores one or more model configs and parameters.
    Parameters are stacked for ensembling.
    """
    if isinstance(model_dir, Path) or isinstance(model_dir, str):
        config, params = restore_single_parameters(model_dir)

    elif isinstance(model_dir, list):
        param_list = []
        for path in model_dir:
            config, params = restore_single_parameters(path)
            param_list.append(params)

        params = stack_parameters(param_list)
    else:
        raise NotImplementedError(
            "Please provide either a path or list of paths to trained models"
        )

    return config, params


def canonicalize_energy_model_parameters(params):
    """Ensures that parameters from EnergyDerivativeModels can be loaded
    into EnergyModels by removing the "energy_model" parameter layer.
    """
    param_dict = unfreeze(params)

    first_level = param_dict["params"]
    if "energy_model" in first_level.keys():
        params = {"params": first_level["energy_model"]}
    params = freeze(params)
    return params


def canonicalize_energy_grad_model_parameters(params):
    """Ensures that parameters from EnergyModels can be loaded
    into EnergyDerivativeModels by adding the "energy_model" parameter layer.
    """
    param_dict = unfreeze(params)

    first_level = param_dict["params"]
    if "energy_model" not in first_level.keys():
        params = {"params": {"energy_model": first_level}}
    params = freeze(params)
    return params



from typing import Callable, Literal, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np
from flax.core.frozen_dict import FrozenDict
from flax.traverse_util import flatten_dict, unflatten_dict
from jax import tree_util
from pydantic import BaseModel, TypeAdapter

from apax.nn.models import EnergyModel

FeatureMap = Callable[[FrozenDict, dict], jax.Array]


class FeatureTransformation(BaseModel):
    def apply(self, model: EnergyModel) -> FeatureMap:
        return model


def extract_feature_params(params: dict, layer_name: str) -> Tuple[dict, dict]:
    """Separate params into those belonging to a selected layer
    and the remaining ones.
    """
    p_flat = flatten_dict(params)

    feature_layer_params = {k: v for k, v in p_flat.items() if layer_name in k}
    remaining_params = {k: v for k, v in p_flat.items() if layer_name not in k}

    if len(feature_layer_params.keys()) > 2:  # w and b
        print(feature_layer_params.keys())
        raise ValueError("Found more than one layer of the specified name")

    return feature_layer_params, remaining_params


class LastLayerGradientFeatures(FeatureTransformation, extra="forbid"):
    """
    Model transfomration which computes the gradient of the output
    wrt. the specified layer.
    https://arxiv.org/pdf/2203.09410

    Parameters
    ----------
    layer_name: str
        Name of the layer wrt. which to take the gradient.
    """

    name: Literal["ll_grad"] = "ll_grad"
    layer_name: str = "dense_2"

    def apply(self, model: EnergyModel) -> FeatureMap:
        def ll_grad(params, inputs):
            ll_params, remaining_params = extract_feature_params(params, self.layer_name)

            def inner(ll_params):
                ll_params.update(remaining_params)
                full_params = unflatten_dict(ll_params)

                # TODO find better abstraction for inputs
                R, Z, idx, box, offsets = (
                    inputs["positions"],
                    inputs["numbers"],
                    inputs["idx"],
                    inputs["box"],
                    inputs["offsets"],
                )
                out = model(full_params, R, Z, idx, box, offsets)
                # take mean in case of shallow ensemble
                # no effect for single model
                out = jnp.mean(out)
                return out

            g_ll = jax.grad(inner)(ll_params)
            g_ll = unflatten_dict(g_ll)
            g_ll = tree_util.tree_map(
                lambda arr: jnp.mean(arr, axis=-1, keepdims=True), g_ll
            )
            g_flat = tree_util.tree_map(lambda arr: jnp.reshape(arr, (-1,)), g_ll)
            (gb, gw), _ = tree_util.tree_flatten(g_flat)

            g = [gw, gb]
            g = jnp.concatenate(g)

            return g

        return ll_grad


class LastLayerForceFeatures(FeatureTransformation, extra="forbid"):
    """
    Model transformation which computes the jacobian of the forces
    wrt. the specified layer.
    For BAL the strategy "flatten" has to be selected.

    Parameters
    ----------
    layer_name: str
        Name of the layer wrt. which to take the jacobian.
    strategy: str
        one of raw, sum, flatten. Only flatten seems to work
        for BAL. raw is required for LLPR.
    """

    name: Literal["ll_force_feat"] = "ll_force_feat"
    layer_name: str = "dense_2"
    strategy: str = "raw"

    def apply(self, model: EnergyModel) -> FeatureMap:
        def ll_grad(params, inputs):
            ll_params, remaining_params = extract_feature_params(params, self.layer_name)

            energy_fn = lambda *inputs: jnp.mean(model(*inputs))
            force_fn = jax.grad(energy_fn, 1)

            def inner(ll_params):
                ll_params.update(remaining_params)
                full_params = unflatten_dict(ll_params)

                R, Z, idx, box, offsets = (
                    inputs["positions"],
                    inputs["numbers"],
                    inputs["idx"],
                    inputs["box"],
                    inputs["offsets"],
                )
                out = force_fn(full_params, R, Z, idx, box, offsets)
                return out

            g_ll = jax.jacfwd(inner)(ll_params)
            g_ll = unflatten_dict(g_ll)

            # shapes:
            # b: n_atoms, 3, 1
            # w: n_atoms, 3, n_features, 1

            if self.strategy == "raw":
                (gb, gw), _ = tree_util.tree_flatten(g_ll)

                # g: n_atoms, 3, n_features
                g = gw[:, :, :, 0]
            elif self.strategy == "sum":
                g_summed = tree_util.tree_map(
                    lambda arr: jnp.reshape(jnp.sum(jnp.sum(arr, 0), 0), (-1,)), g_ll
                )
                (gb, gw), _ = tree_util.tree_flatten(g_summed)
                g = [gw, gb]
                g = jnp.concatenate(g)

            elif self.strategy == "flatten":
                g_flat = tree_util.tree_map(lambda arr: jnp.reshape(arr, (-1,)), g_ll)
                (gb, gw), _ = tree_util.tree_flatten(g_flat)
                g = gw
            else:
                raise ValueError(f"unknown strategy: {self.strategy}")

            return g

        return ll_grad


class FullGradientRPFeatures(FeatureTransformation, extra="forbid"):
    """
    Model transfomration which computes the gradient of the output
    wrt. all parameters and applies a gaussian random projection for
    dimensionality reduction.
    https://arxiv.org/pdf/2203.09410

    Parameters
    ----------
    num_rp: int
        Dimensionality to reduce the features to.
    """

    name: Literal["full_grad_rp"] = "full_grad_rp"
    num_rp: int = 512

    def apply(self, model: EnergyModel) -> FeatureMap:
        def full_grad(params, inputs):
            def inner(params):
                # TODO find better abstraction for inputs
                R, Z, idx, box, offsets = (
                    inputs["positions"],
                    inputs["numbers"],
                    inputs["idx"],
                    inputs["box"],
                    inputs["offsets"],
                )
                out = model(params, R, Z, idx, box, offsets)
                # take mean in case of shallow ensemble
                # no effect for single model
                out = jnp.mean(out)
                return out

            grads = jax.grad(inner)(params)
            grads = tree_util.tree_map(
                lambda arr: jnp.mean(arr, axis=-1, keepdims=True), grads
            )
            g_flat = tree_util.tree_map(lambda arr: jnp.reshape(arr, (-1,)), grads)
            gs, _ = tree_util.tree_flatten(g_flat)
            g = jnp.concatenate(gs)

            with jax.ensure_compile_time_eval():
                n_features = g.shape[0]
                RP = np.random.randn(n_features, self.num_rp) / np.sqrt(self.num_rp)
                RP = jnp.array(RP)

            g_rp = g @ RP

            return g_rp

        return full_grad


class IdentityFeatures(FeatureTransformation, extra="forbid"):
    """Identity feature map. For debugging purposes"""

    name: Literal["identity"]

    def apply(self, model: EnergyModel) -> FeatureMap:
        return model


FeatureMapOptions = TypeAdapter(
    Union[
        LastLayerGradientFeatures,
        LastLayerForceFeatures,
        FullGradientRPFeatures,
        IdentityFeatures,
    ]
).validate_python



import jax.numpy as jnp
import optax
from flax.core.frozen_dict import freeze
from jax import tree_util

from apax.optimizer import get_opt


def test_get_opt():
    params = {
        "my_linear": {"w": jnp.ones((3, 3)), "b": jnp.ones((3,))},
        "my_embed": {
            "atomic_type_embedding": jnp.ones((3, 3)),
        },
        "my_scale_shift": {
            "scale_per_element": jnp.ones((3,)),
            "shift_per_element": jnp.ones((3,)),
        },
    }
    params = freeze(params)

    grads = tree_util.tree_map(lambda x: x * 0.01, tree=params)

    opt = get_opt(
        params,
        0,
        500,
        emb_lr=0.05,
        nn_lr=0.01,
        scale_lr=0.001,
        shift_lr=0.1,
        schedule={"name": "linear", "transition_begin": 0, "end_value": 1e-6},
    )
    opt_state = opt.init(params=params)

    updates, new_opt_state = opt.update(grads, opt_state, params)
    new_params = optax.apply_updates(params, updates)

    filtered_params = tree_util.tree_map(
        lambda v: v[0] if len(v.shape) == 1 else v[0, 0], new_params
    )
    flat_params, treedef = tree_util.tree_flatten(filtered_params)

    assert jnp.allclose(flat_params[1], flat_params[2])
    assert flat_params[-2] > flat_params[1]
    assert flat_params[0] < flat_params[1]
    assert flat_params[-1] < flat_params[1]



import functools
from typing import Any

import e3x
import flax.linen as nn
import jax
import jax.numpy as jnp

from apax.layers.descriptor.basis_functions import BesselBasis
from apax.layers.masking import mask_by_neighbor
from apax.utils.convert import str_to_dtype


class EquivMPRepresentation(nn.Module):
    basis_fn: nn.Module = BesselBasis()
    features: int = 32
    max_degree: int = 2
    num_iterations: int = 3
    max_atomic_number: int = 118
    dtype: Any = jnp.float32
    apply_mask: bool = True

    @nn.compact
    def __call__(self, dr_vec, Z, neighbor_idxs):
        dtype = str_to_dtype(self.dtype)
        dr_vec = dr_vec.astype(dtype)

        idx_i, idx_j = neighbor_idxs[0], neighbor_idxs[1]
        # 1. Calculate displacement vectors.
        displacements = dr_vec  # Shape (num_pairs, 3).

        # 2. Expand displacement vectors in basis functions.
        basis = (
            e3x.nn.basis(  # Shape (num_pairs, 1, (max_degree+1)**2, num_basis_functions).
                displacements,
                num=self.basis_fn.n_basis,
                max_degree=self.max_degree,
                radial_fn=lambda x, y: self.basis_fn(x),
                cutoff_fn=functools.partial(
                    e3x.nn.smooth_cutoff, cutoff=self.basis_fn.r_max
                ),
            )
        )

        if self.apply_mask:
            basis = mask_by_neighbor(basis, neighbor_idxs)

        # 3. Embed atomic numbers in feature space, x has shape (num_atoms, 1, 1, features).
        x = e3x.nn.Embed(
            num_embeddings=self.max_atomic_number + 1, features=self.features
        )(Z)
        x = x.astype(dtype)

        # 4. Perform iterations (message-passing + atom-wise refinement).
        for i in range(self.num_iterations):
            # Message-pass.
            if i == self.num_iterations - 1:  # Final iteration.
                # Since we will only use scalar features after the final message-pass, we do not want to produce non-scalar
                # features for efficiency reasons.
                y = e3x.nn.MessagePass(max_degree=0, include_pseudotensors=False)(
                    x, basis, dst_idx=idx_i, src_idx=idx_j
                )
                # After the final message pass, we can safely throw away all non-scalar features.
                x = e3x.nn.change_max_degree_or_type(
                    x, max_degree=0, include_pseudotensors=False
                )
            else:
                # In intermediate iterations, the message-pass should consider all possible coupling paths.
                y = e3x.nn.MessagePass()(x, basis, dst_idx=idx_i, src_idx=idx_j)

            x = x.astype(dtype)
            y = y.astype(dtype)
            y = e3x.nn.add(x, y)

            # Atom-wise refinement MLP.
            y = e3x.nn.Dense(self.features)(y)
            y = e3x.nn.silu(y)
            y = e3x.nn.Dense(self.features, kernel_init=jax.nn.initializers.zeros)(y)

            # Residual connection.
            x = e3x.nn.add(x, y)

        x = x[:, 0, 0, :]
        return x



from functools import partial
from pathlib import Path
from typing import Callable, Union

import ase
import jax
import jax.numpy as jnp
import numpy as np
from ase.calculators.calculator import Calculator, all_changes
from ase.calculators.singlepoint import SinglePointCalculator
from flax.core.frozen_dict import freeze, unfreeze
from jax import tree_util
from tqdm import trange
from vesin import NeighborList

from apax.data.input_pipeline import (
    CachedInMemoryDataset,
)
from apax.train.checkpoints import check_for_ensemble, restore_parameters
from apax.utils.jax_md_reduced import partition, quantity, space


def maybe_vmap(apply, params):
    n_models = check_for_ensemble(params)

    if n_models > 1:
        apply = jax.vmap(apply, in_axes=(0, None, None, None, None, None))

    # Maybe the partial mapping should happen at the very end of initialize
    # That way other functions can mae use of the parameter shape information
    energy_fn = partial(apply, params)
    return energy_fn


def build_energy_neighbor_fns(atoms, config, params, dr_threshold, neigbor_from_jax):
    r_max = config.model.basis.r_max
    box = jnp.asarray(atoms.cell.array, dtype=jnp.float64)
    neigbor_from_jax = neighbor_calculable_with_jax(box, r_max)
    box = box.T
    displacement_fn = None
    neighbor_fn = None

    if neigbor_from_jax:
        if np.all(box < 1e-6):
            displacement_fn, _ = space.free()
        else:
            displacement_fn, _ = space.periodic_general(box, fractional_coordinates=True)

        neighbor_fn = partition.neighbor_list(
            displacement_fn,
            box,
            config.model.basis.r_max,
            dr_threshold,
            fractional_coordinates=True,
            disable_cell_list=True,
            format=partition.Sparse,
        )

    n_species = 119  # int(np.max(Z) + 1)
    Builder = config.model.get_builder()
    builder = Builder(config.model.model_dump(), n_species=n_species)

    model = builder.build_energy_derivative_model(
        apply_mask=True, init_box=np.array(box), inference_disp_fn=displacement_fn
    )

    energy_fn = maybe_vmap(model.apply, params)
    return energy_fn, neighbor_fn


def process_stress(results, box):
    V = quantity.volume(3, box)
    results = {
        # We should properly check whether CP2K uses the ASE cell convention
        # for tetragonal strain, it doesn't matter whether we transpose or not
        k: val.T / V if k.startswith("stress") else val
        for k, val in results.items()
    }
    return results


def make_ensemble(model):
    def ensemble(positions, Z, idx, box, offsets):
        results = model(positions, Z, idx, box, offsets)
        uncertainty = {k + "_uncertainty": jnp.std(v, axis=0) for k, v in results.items()}
        ensemble = {k + "_ensemble": v for k, v in results.items()}
        results = {k: jnp.mean(v, axis=0) for k, v in results.items()}
        if "forces_ensemble" in ensemble.keys():
            ensemble["forces_ensemble"] = jnp.transpose(
                ensemble["forces_ensemble"], (1, 2, 0)
            )
        if "forces_ensemble" in ensemble.keys():
            ensemble["stress_ensemble"] = jnp.transpose(
                ensemble["forces_ensemble"], (1, 2, 0)
            )
        results.update(uncertainty)
        results.update(ensemble)

        return results

    return ensemble


def unpack_results(results, inputs):
    n_structures = len(results["energy"])
    unpacked_results = []
    for i in range(n_structures):
        single_results = tree_util.tree_map(lambda x: x[i], results)
        single_results["energy"] = single_results["energy"].item()
        for k, v in single_results.items():
            if "forces" in k:
                single_results[k] = v[: inputs["n_atoms"][i]]
        unpacked_results.append(single_results)
    return unpacked_results


class ASECalculator(Calculator):
    """
    ASE Calculator for apax models.
    Always implements energy and force predictions.
    Stress predictions and corresponding uncertainties are added to
    `implemented_properties` based on whether the stress flag is set
    in the model config and whether a model ensemble is loaded.

    """

    implemented_properties = [
        "energy",
        "forces",
    ]

    def __init__(
        self,
        model_dir: Union[Path, list[Path]],
        dr_threshold: float = 0.5,
        transformations: Callable = [],
        padding_factor: float = 1.5,
        **kwargs,
    ):
        """
        Parameters
        ----------
        model_dir:
            Path to a model directory of the form `.../directory/experiment`
            (see Config docs for details).
            If a list of model paths is provided, they will be ensembled.
        dr_threshold:
            Neighborlist skin for the JaxMD neighborlist.
        transformations:
            Function transformations applied on top of the EnergyDerivativeModel.
            Transfomrations are implemented under `apax.md.transformations`.
        padding_factor:
            Multiple of the fallback vesin's amount of neighbors.
            This NL will be padded to `len(neighbors) * padding_factor`
            on NL initialization.
        """
        Calculator.__init__(self, **kwargs)
        self.dr_threshold = dr_threshold
        self.transformations = transformations

        self.model_config, self.params = restore_parameters(model_dir)
        self.n_models = check_for_ensemble(self.params)
        self.padding_factor = padding_factor
        self.padded_length = 0

        if self.model_config.model.calc_stress:
            self.implemented_properties.append("stress")

        if self.n_models > 1:
            uncertainty_kws = [
                prop + "_uncertainty" for prop in self.implemented_properties
            ]
            self.implemented_properties += uncertainty_kws

        self.step = None
        self.neighbor_fn = None
        self.neighbors = None
        self.offsets = None
        self.model = None

    def initialize(self, atoms):
        box = jnp.asarray(atoms.cell.array, dtype=jnp.float64)
        self.r_max = self.model_config.model.basis.r_max
        self.neigbor_from_jax = neighbor_calculable_with_jax(box, self.r_max)
        model, neighbor_fn = build_energy_neighbor_fns(
            atoms,
            self.model_config,
            self.params,
            self.dr_threshold,
            self.neigbor_from_jax,
        )

        if self.n_models > 1:
            model = make_ensemble(model)

        for transformation in self.transformations:
            model = transformation.apply(model)

        self.model = model
        self.step = get_step_fn(model, atoms, self.neigbor_from_jax)
        self.neighbor_fn = neighbor_fn

        if self.neigbor_from_jax:
            positions = jnp.asarray(atoms.positions, dtype=jnp.float64)
            if np.any(atoms.get_cell().lengths() > 1e-6):
                box = atoms.cell.array.T
                inv_box = jnp.linalg.inv(box)
                positions = space.transform(inv_box, positions)  # frac coords
                self.neighbors = self.neighbor_fn.allocate(positions, box=box)
            else:
                self.neighbors = self.neighbor_fn.allocate(positions)
        else:
            calculator = NeighborList(cutoff=self.r_max, full_list=True)
            idxs_i, _, _ = calculator.compute(
                points=atoms.positions,
                box=atoms.cell.array,
                periodic=np.any(atoms.pbc),
                quantities="ijS",
            )
            self.padded_length = int(len(idxs_i) * self.padding_factor)

    def set_neighbours_and_offsets(self, atoms, box):
        calculator = NeighborList(cutoff=self.r_max, full_list=True)
        idxs_i, idxs_j, offsets = calculator.compute(
            points=atoms.positions,
            box=atoms.cell.array,
            periodic=np.any(atoms.pbc),
            quantities="ijS",
        )
        if len(idxs_i) > self.padded_length:
            print("neighbor list overflowed, extending.")
            self.initialize(atoms)

        zeros_to_add = self.padded_length - len(idxs_i)

        self.neighbors = np.array([idxs_i, idxs_j], dtype=np.int32)
        self.neighbors = np.pad(self.neighbors, ((0, 0), (0, zeros_to_add)), "constant")

        offsets = np.matmul(offsets, box)
        self.offsets = np.pad(offsets, ((0, zeros_to_add), (0, 0)), "constant")

    def calculate(self, atoms, properties=["energy"], system_changes=all_changes):
        Calculator.calculate(self, atoms, properties, system_changes)
        positions = jnp.asarray(atoms.positions, dtype=jnp.float64)
        box = jnp.asarray(atoms.cell.array, dtype=jnp.float64)

        # setup model and neighbours
        if self.step is None:
            self.initialize(atoms)

        elif "numbers" in system_changes:
            self.initialize(atoms)

            if self.neigbor_from_jax:
                self.neighbors = self.neighbor_fn.allocate(positions)

        elif "cell" in system_changes:
            neigbor_from_jax = neighbor_calculable_with_jax(box, self.r_max)
            if self.neigbor_from_jax != neigbor_from_jax:
                self.initialize(atoms)

        # predict
        if self.neigbor_from_jax:
            results, self.neighbors = self.step(positions, self.neighbors, box)

            if self.neighbors.did_buffer_overflow:
                print("neighbor list overflowed, reallocating.")
                self.initialize(atoms)
                results, self.neighbors = self.step(positions, self.neighbors, box)

        else:
            self.set_neighbours_and_offsets(atoms, box)
            positions = np.array(space.transform(np.linalg.inv(box), atoms.positions))

            results = self.step(positions, self.neighbors, box, self.offsets)

        self.results = {k: np.array(v, dtype=np.float64) for k, v in results.items()}
        self.results["energy"] = self.results["energy"].item()

    def batch_eval(
        self, atoms_list: list[ase.Atoms], batch_size: int = 64, silent: bool = False
    ) -> list[ase.Atoms]:
        """Evaluate the model on a list of Atoms. This is preferable to assigning
        the calculator to each Atoms instance for 2 reasons:
        1. Processing can be abtched, which is advantageous for larger datasets.
        2. Inputs are padded so no recompilation is triggered when evaluating
        differently sized systems.

        Parameters
        ----------
        atoms_list :
            List of Atoms to be evaluated.
        batch_size:
            Processing batch size. Does not affect results,
            only speed and memory requirements.
        silent:
            Whether or not to suppress progress bars.

        Returns
        -------
        evaluated_atoms_list:
            List of Atoms with labels predicted by the model.
        """

        init_box = atoms_list[0].cell.array

        dataset = CachedInMemoryDataset(
            atoms_list,
            self.model_config.model.basis.r_max,
            batch_size,
            n_epochs=1,
            ignore_labels=True,
        )

        evaluated_atoms_list = []
        n_data = dataset.n_data
        ds = dataset.batch()

        Builder = self.model_config.model.get_builder()
        builder = Builder(self.model_config.model.model_dump())
        model = builder.build_energy_derivative_model(
            apply_mask=True,
            init_box=init_box,
        ).apply

        if self.n_models > 1:
            model = jax.vmap(model, in_axes=(0, None, None, None, None, None))

        model = partial(model, self.params)

        if self.n_models > 1:
            model = make_ensemble(model)

        for transformation in self.transformations:
            model = transformation.apply(model)

        model = jax.vmap(model, in_axes=(0, 0, 0, 0, 0))
        model = jax.jit(model)

        pbar = trange(
            n_data, desc="Evaluating data", ncols=100, leave=False, disable=silent
        )
        for i, inputs in enumerate(ds):
            results = model(
                inputs["positions"],
                inputs["numbers"],
                inputs["idx"],
                inputs["box"],
                inputs["offsets"],
            )
            unpadded_results = unpack_results(results, inputs)

            # for the last batch, the number of structures may be less
            # than the batch_size,  which is why we check this explicitly
            num_strucutres_in_batch = results["energy"].shape[0]
            for j in range(num_strucutres_in_batch):
                atoms = atoms_list[i * batch_size + j].copy()
                atoms.calc = SinglePointCalculator(atoms=atoms, **unpadded_results[j])
                evaluated_atoms_list.append(atoms)
            pbar.update(batch_size)
        pbar.close()
        dataset.cleanup()
        return evaluated_atoms_list

    @property
    def ll_weights(self):
        dense_layers = list(
            self.params["params"]["energy_model"]["atomistic_model"]["readout"].keys()
        )
        llweights = self.params["params"]["energy_model"]["atomistic_model"]["readout"][
            dense_layers[-1]
        ]["w"]
        return np.asarray(llweights)

    def set_ll_weights(self, new_weights):
        params = unfreeze(self.params)
        dense_layers = list(
            params["params"]["energy_model"]["atomistic_model"]["readout"].keys()
        )
        params["params"]["energy_model"]["atomistic_model"]["readout"][dense_layers[-1]][
            "w"
        ] = jnp.asarray(new_weights, dtype=jnp.float32)
        self.params = freeze(params)
        self.step = None


def neighbor_calculable_with_jax(box, r_max):
    if np.all(box < 1e-6):
        return True
    else:
        # all lettice vector combinations to calculate all three plane distances
        a_vec_list = [box[0], box[0], box[1]]
        b_vec_list = [box[1], box[2], box[2]]
        c_vec_list = [box[2], box[1], box[0]]

        height = []
        for i in range(3):
            normvec = np.cross(a_vec_list[i], b_vec_list[i])
            projection = (
                c_vec_list[i]
                - np.sum(normvec * c_vec_list[i]) / np.sum(normvec**2) * normvec
            )
            height.append(np.linalg.norm(c_vec_list[i] - projection))

        if np.min(height) / 2 > r_max:
            return True
        else:
            return False


def get_step_fn(model, atoms, neigbor_from_jax):
    Z = jnp.asarray(atoms.numbers)
    if neigbor_from_jax:

        @jax.jit
        def step_fn(positions, neighbor, box):
            if np.any(atoms.get_cell().lengths() > 1e-6):
                box = box.T
                inv_box = jnp.linalg.inv(box)
                positions = space.transform(inv_box, positions)
                neighbor = neighbor.update(positions, box=box)
            else:
                neighbor = neighbor.update(positions)

            offsets = jnp.full([neighbor.idx.shape[1], 3], 0)
            results = model(positions, Z, neighbor.idx, box, offsets)

            if "stress" in results.keys():
                results = process_stress(results, box)

            return results, neighbor

    else:

        @jax.jit
        def step_fn(positions, neighbor, box, offsets):
            results = model(positions, Z, neighbor, box, offsets)
            if "stress" in results.keys():
                results = process_stress(results, box)
            return results

    return step_fn



import jax
import jax.numpy as jnp
import pytest

from apax.layers.properties import PropertyHead
from apax.layers.readout import AtomisticReadout


@pytest.fixture
def setup_data():
    """Fixture to provide dummy data for testing."""
    n_atoms = 5
    n_features = 3
    n_species = 119

    g = jnp.ones((n_atoms, n_features))
    R = jnp.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 1.0, 0.0],
            [0.0, 1.0, 1.0],
        ]
    )  # atom positions
    dr_vec = None
    Z = jnp.array([1, 6, 8, 1, 6])  # atomic numbers
    idx = None
    box = None

    return g, R, dr_vec, Z, idx, box


def test_property_head(setup_data):
    """Test PropertyHead class functionality."""
    g, R, dr_vec, Z, idx, box = setup_data

    # Instantiate PropertyHead
    property_head = PropertyHead(pname="property")

    # Test setup
    params = property_head.init(jax.random.PRNGKey(0), g, R, dr_vec, Z, idx, box)
    assert "scale_per_element" in params["params"]
    assert "shift_per_element" in params["params"]

    # Test forward pass for mode `l0`
    property_head = PropertyHead(
        pname="property", mode="l0", apply_mask=False, aggregation="none"
    )
    output = property_head.apply(params, g, R, dr_vec, Z, idx, box)
    assert "property" in output.keys()
    assert output["property"].shape == (5, 1)  # Shape should match input atoms x features

    # Test aggregation: sum
    property_head = PropertyHead(
        pname="property", mode="l0", apply_mask=False, aggregation="sum"
    )
    output = property_head.apply(params, g, R, dr_vec, Z, idx, box)
    assert "property" in output.keys()
    assert output["property"].shape == (1,)

    # Test mode `l1`
    property_head = PropertyHead(pname="property", mode="l1", apply_mask=True)
    output = property_head.apply(params, g, R, dr_vec, Z, idx, box)
    assert "property" in output.keys()
    assert output["property"].shape == (5, 3)

    # Test mode `symmetric_traceless_l2`
    property_head = PropertyHead(
        pname="property", mode="symmetric_traceless_l2", apply_mask=True
    )
    output = property_head.apply(params, g, R, dr_vec, Z, idx, box)
    assert "property" in output.keys()
    assert output["property"].shape == (5, 3, 3)

    # Test ensemble detection
    property_head = PropertyHead(
        pname="property",
        readout=AtomisticReadout(n_shallow_ensemble=10),
        mode="l0",
        apply_mask=False,
        aggregation="mean",
    )
    params = property_head.init(jax.random.PRNGKey(0), g, R, dr_vec, Z, idx, box)
    output = property_head.apply(params, g, R, dr_vec, Z, idx, box)
    assert "property" in output.keys()
    assert "property_uncertainty" in output



import jax
import jax.numpy as jnp

from apax.layers.descriptor.basis_functions import GaussianBasis, RadialFunction


def test_gaussian_basis():
    n_basis = 5
    key = jax.random.PRNGKey(0)
    basis = GaussianBasis(n_basis=n_basis)

    dr = jnp.array([0.5, 1.0, 2.0, 3.0])  # n_neighbors

    params = basis.init(key, dr)
    assert len(params.keys()) == 0
    result = basis.apply(params, dr)
    assert result.shape == (4, n_basis)  # n_neighbors x n_basis


def test_radial_function():
    key = jax.random.PRNGKey(0)
    n_species = 119  # 3
    n_basis = 5
    n_radial = 2

    dr = jnp.array([0.5, 1.0, 2.0, 3.0], dtype=jnp.float32)  # n_neighbors
    Z_i = jnp.array([1, 2, 1, 2])
    Z_j = jnp.array([2, 1, 2, 1])
    # cutoff = jnp.array([1.0, 1.0, 1.0, 1.0], dtype=jnp.float32)

    radial_fn = RadialFunction(
        n_species=n_species, n_radial=n_radial, basis_fn=GaussianBasis(n_basis)
    )

    params = radial_fn.init(key, dr, Z_i, Z_j)
    result = radial_fn.apply(params, dr, Z_i, Z_j)

    assert params["params"]["atomic_type_embedding"].shape == (
        n_species,
        n_species,
        n_radial,
        n_basis,
    )
    assert result.shape == (4, n_radial)  # n_neighbors x n_radial



import jax.numpy as jnp
import numpy as np

from apax.train.loss import (
    Loss,
    force_angle_div_force_label,
    force_angle_exponential_weight,
    force_angle_loss,
    weighted_squared_error,
)


def test_weighted_squared_error():
    name = "energy"
    label = {"energy": jnp.array([[0.1, 0.4, 0.2, -0.5], [0.1, -0.1, 0.8, 0.6]])}

    loss = weighted_squared_error(label, label, name)
    loss = jnp.sum(loss)
    ref = 0.0
    assert loss.shape == ()
    assert abs(loss - ref) < 1e-6

    pred = {
        "energy": jnp.array(
            [
                [0.6, 0.4, 0.2, -0.5],
                [0.1, -0.1, 0.8, 0.6],
            ]
        )
    }
    loss = weighted_squared_error(label, pred, name)
    loss = jnp.sum(loss)
    ref = 0.25
    assert abs(loss - ref) < 1e-6


def test_force_angle_loss():
    name = "forces"
    F_pred = {
        "forces": jnp.array(
            [
                [
                    [0.5, 0.0, 0.0],
                    [0.5, 0.0, 0.0],
                    [0.5, 0.5, 0.0],
                    [0.0, 0.5, 0.0],
                    [0.0, 0.5, 0.0],
                    [0.0, 0.0, 0.0],  # padding
                ]
            ]
        )
    }

    F_0 = {
        "forces": jnp.array(
            [
                [
                    [0.5, 0.0, 0.0],
                    [0.9, 0.0, 0.0],
                    [0.5, 0.0, 0.0],
                    [0.5, 0.0, 0.0],
                    [0.9, 0.0, 0.0],
                    [0.0, 0.0, 0.0],  # padding
                ]
            ]
        )
    }

    F_angle_loss = force_angle_loss(F_pred, F_0, name)
    F_angle_loss = jnp.arccos(-F_angle_loss + 1) * 360 / (2 * np.pi)
    assert F_angle_loss.shape == (1, 6)
    ref = jnp.array([0.0, 0.0, 45.0, 90.0, 90.0, 90.0])
    assert jnp.allclose(F_angle_loss, ref)

    F_angle_loss = force_angle_div_force_label(F_pred, F_0, name)
    assert F_angle_loss.shape == (1, 6)

    F_angle_loss = force_angle_exponential_weight(F_pred, F_0, name)
    assert F_angle_loss.shape == (1, 6)


def test_force_loss():
    name = "forces"
    loss_type = "mse"
    weight = 1
    inputs = {
        "n_atoms": jnp.array([2]),
    }
    label = {
        name: jnp.array(
            [
                [
                    [0.4, 0.2, 0.5],
                    [0.3, 0.8, 0.1],
                ]
            ]
        ),
    }

    pred = {
        name: jnp.array(
            [
                [
                    [0.4, 0.2, 0.5],
                    [0.3, 0.8, 0.1],
                ]
            ]
        ),
    }
    loss_func = Loss(name=name, loss_type=loss_type, weight=weight)
    loss = loss_func(inputs=inputs, label=label, prediction=pred)
    ref_loss = 0.0
    assert loss.shape == ()
    assert abs(loss - ref_loss) < 1e-6

    pred = {
        name: jnp.array(
            [
                [
                    [0.4, 0.2, 0.5],
                    [0.3, 0.8, 0.6],
                ]
            ]
        ),
    }
    loss_func = Loss(name=name, loss_type=loss_type, weight=weight)
    loss = loss_func(inputs=inputs, label=label, prediction=pred)
    ref_loss = 0.125
    assert abs(loss - ref_loss) < 1e-6

    weight = 2
    loss_func = Loss(name=name, loss_type=loss_type, weight=weight)
    loss = loss_func(inputs=inputs, label=label, prediction=pred)
    ref_loss = 0.25
    assert abs(loss - ref_loss) < 1e-6



import logging
from typing import Literal, Union

import jax.numpy as jnp
from pydantic import BaseModel, TypeAdapter

log = logging.getLogger(__name__)


class DynamicsCheckBase(BaseModel):
    def check(self, predictions, positions, box):
        pass


class EnergyUncertaintyCheck(DynamicsCheckBase, extra="forbid"):
    name: Literal["energy_uncertainty"] = "energy_uncertainty"
    threshold: float
    per_atom: bool = True

    def check(self, predictions, positions, box):
        if "energy_uncertainty" not in predictions.keys():
            m = "No energy uncertainty found. Are you using a model ensemble?"
            raise ValueError(m)

        energy_uncertainty = predictions["energy_uncertainty"]
        if self.per_atom:
            n_atoms = predictions["forces"].shape[0]
            energy_uncertainty = energy_uncertainty / n_atoms

        check_passed = jnp.all(energy_uncertainty < self.threshold)
        return check_passed


class ForceUncertaintyCheck(DynamicsCheckBase, extra="forbid"):
    name: Literal["forces_uncertainty"] = "forces_uncertainty"
    threshold: float

    def check(self, predictions, positions, box):
        if "forces_uncertainty" not in predictions.keys():
            m = "No force uncertainties found. Are you using a model ensemble?"
            raise ValueError(m)

        forces_uncertainty = predictions["forces_uncertainty"]

        check_passed = jnp.all(forces_uncertainty < self.threshold)
        return check_passed


class ReflectionCheck(DynamicsCheckBase, extra="forbid"):
    name: Literal["reflection"] = "reflection"
    cutoff_plane_height: float

    def check(self, predictions, positions, box):
        cartesian = positions @ box
        z_pos = cartesian[:, 2]

        check_passed = jnp.all(z_pos < self.cutoff_plane_height)

        return check_passed


DynamicsChecks = TypeAdapter(
    Union[EnergyUncertaintyCheck, ForceUncertaintyCheck, ReflectionCheck]
).validate_python



from typing import Any

import jax.numpy as jnp
from jax import random
from jax._src import dtypes
from jax.nn.initializers import Initializer

Array = Any
KeyArray = Array

DTypeLikeFloat = Any
DTypeLikeComplex = Any
DTypeLikeInexact = Any
RealNumeric = Any


def uniform_range(minval, maxval, dtype: DTypeLikeInexact = jnp.float_) -> Initializer:
    """Builds an initializer that returns real uniformly-distributed random arrays
    in a specified range.
    """

    def init(key: KeyArray, shape, dtype: DTypeLikeInexact = dtype) -> Array:
        dtype = dtypes.canonicalize_dtype(dtype)
        return random.uniform(key, shape, dtype, minval, maxval)

    return init



from typing import Literal

from pydantic import BaseModel, NonNegativeFloat


class LRSchedule(BaseModel, frozen=True, extra="forbid"):
    name: str


class LinearLR(LRSchedule, frozen=True, extra="forbid"):
    """
    Configuration of the optimizer.
    Learning rates of 0 will freeze the respective parameters.

    Parameters
    ----------
    name : str, default = "adam"
    transition_begin: int = 0
        Number of steps after which to start decreasing
    end_value: NonNegativeFloat = 1e-6
        Final LR at the end of training.
    """

    name: Literal["linear"]
    transition_begin: int = 0
    end_value: NonNegativeFloat = 1e-6


class CyclicCosineLR(LRSchedule, frozen=True, extra="forbid"):
    """
    Configuration of the optimizer.
    Learning rates of 0 will freeze the respective parameters.

    Parameters
    ----------
    period: int = 40
        Length of a cycle in epochs.
    decay_factor: NonNegativeFloat = 0.93
        Factor by which to decrease the LR after each cycle.
        1.0 means no decrease.
    """

    name: Literal["cyclic_cosine"]
    period: int = 40
    decay_factor: NonNegativeFloat = 0.93



import pathlib
import uuid

import jax
import numpy as np
import pytest
from ase.io import write

from apax.train.checkpoints import restore_parameters
from tests.conftest import load_config_and_run_training

TEST_PATH = pathlib.Path(__file__).parent.resolve()


def l2_param_diff(p1, p2):
    p1, _ = jax.tree.flatten(p1)
    p2, _ = jax.tree.flatten(p2)
    diff = 0.0
    for i in range(len(p1)):
        diff += np.sum((p1[i] - p2[i]) ** 2)
    return diff


@pytest.mark.parametrize("num_data", (30,))
def test_transfer_learning(get_tmp_path, example_dataset):
    config_path = TEST_PATH / "config_base.yaml"
    config_ft_path = TEST_PATH / "config_ft.yaml"
    working_dir = get_tmp_path / str(uuid.uuid4())
    data_path = get_tmp_path / "ds.extxyz"

    write(data_path, example_dataset)

    data_config_mods = {
        "data": {
            "directory": working_dir.as_posix(),
            "experiment": "base",
            "data_path": data_path.as_posix(),
        },
    }
    load_config_and_run_training(config_path, data_config_mods)

    data_config_mods = {
        "data": {
            "directory": working_dir.as_posix(),
            "experiment": "fine_tune",
            "data_path": data_path.as_posix(),
        },
        "checkpoints": {"base_model_checkpoint": (working_dir / "base").as_posix()},
    }
    load_config_and_run_training(config_ft_path, data_config_mods)

    data_config_mods = {
        "data": {
            "directory": working_dir.as_posix(),
            "experiment": "fine_tune_no_pre_training",
            "data_path": data_path.as_posix(),
        },
    }
    load_config_and_run_training(config_ft_path, data_config_mods)

    # Compare parameters
    _, base_params = restore_parameters(working_dir / "base")
    _, ft_params = restore_parameters(working_dir / "fine_tune")
    _, ft_no_pre_params = restore_parameters(working_dir / "fine_tune_no_pre_training")

    diff_base_ft = l2_param_diff(base_params, ft_params)
    diff_base_no_pre = l2_param_diff(base_params, ft_no_pre_params)

    assert diff_base_ft < diff_base_no_pre



import jax
import jax.numpy as jnp

from apax.layers.scaling import PerElementScaleShift


def test_per_element_scale_shift():
    key = jax.random.PRNGKey(0)

    x = jnp.array([1.0, 1.0, 1.0])[:, None]
    Z = jnp.array([1, 2, 2])
    n_species = 119  # 3

    global_shift = 2.0
    global_scale = 1.0

    scale_shift = PerElementScaleShift(
        n_species=n_species, scale=global_scale, shift=global_shift
    )

    params = scale_shift.init(key, x, Z)
    result = scale_shift.apply(params, x, Z)

    assert params["params"]["scale_per_element"].shape == (n_species, 1)
    assert params["params"]["shift_per_element"].shape == (n_species, 1)

    assert jnp.allclose(result, jnp.array([3.0, 3.0, 3.0])[:, None])

    indiv_scale = jnp.array([10.0, 2.0, 3.0])
    indiv_shift = jnp.array([0.0, -2.0, 2.0])
    scale_shift = PerElementScaleShift(
        n_species=n_species, scale=indiv_scale, shift=indiv_shift
    )

    params = scale_shift.init(key, x, Z)
    result = scale_shift.apply(params, x, Z)

    assert params["params"]["scale_per_element"].shape == (3, 1)
    assert params["params"]["shift_per_element"].shape == (3, 1)

    assert jnp.allclose(result, jnp.array([0.0, 5.0, 5.0])[:, None])



import os
import pathlib
import shutil
import subprocess

import zntrack

from apax.nodes.md import ApaxJaxMD
from apax.nodes.model import Apax
from apax.nodes.utils import AddData

CONFIG_PATH = pathlib.Path(__file__).parent / "example.yaml"
MD_CONFIG_PATH = pathlib.Path(__file__).parent / "md.yaml"


def test_n_jax_md(tmp_path, get_md22_stachyose):
    os.chdir(tmp_path)
    subprocess.run(["git", "init"], check=True)
    subprocess.run(["dvc", "init"], check=True)
    shutil.copy(CONFIG_PATH, tmp_path / "example.yaml")
    shutil.copy(MD_CONFIG_PATH, tmp_path / "md.yaml")
    proj = zntrack.Project()
    with proj:
        data = AddData(file=get_md22_stachyose, stop=100)
        model = Apax(data=data.frames, validation_data=data.frames, config="example.yaml")
        md = ApaxJaxMD(model=model, config="md.yaml", data=data.frames)

    proj.repro()

    md = md.from_rev()
    assert len(md.frames) == 50



import jax.numpy as jnp

from apax.config.train_config import MetricsConfig
from apax.train.metrics import cosine_sim, initialize_metrics


def test_cosine_sim():
    prediction = {
        "forces": jnp.array(
            [
                [
                    [0.5, 0.0, 0.0],
                    [0.5, 0.0, 0.0],
                ]
            ]
        )
    }

    label = {
        "forces": jnp.array(
            [
                [
                    [0.0, 0.5, 0.0],
                    [0.5, 0.0, 0.0],
                ]
            ]
        )
    }

    angle_error = cosine_sim(label, prediction, "forces")
    assert angle_error.shape == ()
    ref = 0.5
    assert abs(angle_error - ref) < 1e-6


def test_initialize_metrics_collection():
    prediction = {
        "energy": jnp.array([1.0, 1.0]),
        "forces": jnp.array(
            [
                [[1.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
                [[1.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
            ]
        ),
    }

    label = {
        "energy": jnp.array([1.0, 2.0]),
        "forces": jnp.array(
            [
                [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]],
                [[1.0, 0.0, 0.0], [2.0, 1.0, 0.0]],
            ]
        ),
    }
    metrics_list = [
        MetricsConfig(name="energy", reductions=["mae", "rmse"]),
        MetricsConfig(name="forces", reductions=["mae", "mse"]),
    ]
    Metrics = initialize_metrics(metrics_list)
    batch_metrics = Metrics.single_from_model_output(label=label, prediction=prediction)

    epoch_metrics = batch_metrics.compute()

    assert abs(epoch_metrics["energy_mae"] - 0.5) < 1e-6
    assert abs(epoch_metrics["energy_rmse"] - jnp.sqrt(0.5)) < 1e-6
    assert abs(epoch_metrics["forces_mae"] - 1 / 3) < 1e-6
    assert abs(epoch_metrics["forces_mse"] - 1 / 3) < 1e-6



import logging

import ase.io
import zntrack


class AddData(zntrack.Node):
    file: str = zntrack.deps_path()
    start: int = zntrack.params(0)
    stop: int = zntrack.params(None)
    step: int = zntrack.params(1)

    def run(self):
        pass

    @property
    def frames(self) -> list[ase.Atoms]:
        with self.state.fs.open(self.file, "r") as f:
            return ase.io.read(f, index=slice(self.start, self.stop, self.step))


def check_duplicate_keys(dict_a: dict, dict_b: dict, log: logging.Logger) -> None:
    """Check if a key of dict_a is present in dict_b and then log a warning."""
    for key in dict_a:
        if key in dict_b:
            log.warning(
                f"Found <{key}> in given config file. Please be aware that <{key}>"
                " will be overwritten by MLSuite!"
            )



import logging
import typing as t
from pathlib import Path

from flax.training import checkpoints

log = logging.getLogger(__name__)


def load_md_state(state: t.Any, ckpt_dir: Path) -> tuple[t.Any, int]:
    try:
        log.info(f"loading MD state from {ckpt_dir}")
        target = {"state": state, "step": 0}
        restored_ckpt = checkpoints.restore_checkpoint(ckpt_dir, target=target, step=None)
    except FileNotFoundError:
        print(f"No checkpoint found at {ckpt_dir}")
    state = restored_ckpt["state"]
    step = restored_ckpt["step"]
    return state, step



import flax.linen as nn
import jax
import jax.numpy as jnp
from jax import Array

from apax.layers.masking import mask_by_atom
from apax.layers.readout import AtomisticReadout
from apax.utils.math import fp64_sum


def stress_times_vol(energy_fn, position: Array, box, **kwargs) -> Array:
    """Computes the internal stress of a system multiplied with the box volume.
    For training purposes.

    Parameters
    ----------
    energy_fn:
        A function that computes the energy of the system. This
        function must take as an argument `perturbation` which perturbs the
        box shape. Any energy function constructed using `smap` or in `energy.py`
        with a standard space will satisfy this property.
    position:
        An array of particle positions.
    box:
        A box specifying the shape of the simulation volume. Used to infer the
        volume of the unit cell.

    Returns
    -------
    Array
        A float specifying the stress of the system.
    """
    dim = position.shape[1]
    zero = jnp.zeros((dim, dim), position.dtype)
    zero = 0.5 * (zero + zero.T)
    identity = jnp.eye(dim, dtype=position.dtype)

    def U(eps):
        return energy_fn(position, box=box, perturbation=(identity + eps), **kwargs)

    dUdV = jax.grad(U)
    return dUdV(zero)


class PropertyHead(nn.Module):
    """
    the readout is currently limited to a single number
    """

    pname: str
    readout: nn.Module = AtomisticReadout()
    aggregation: str = "none"
    mode: str = "l0"
    apply_mask: bool = True

    def setup(self):
        n_species = 119
        scale_init = nn.initializers.constant(1.0)
        self.scale = self.param(
            "scale_per_element", scale_init, (n_species, 1), jnp.float64
        )

        shift_init = nn.initializers.constant(0.0)

        self.shift_param = self.param(
            "shift_per_element", shift_init, (n_species, 1), jnp.float64
        )

    def __call__(self, g, R, dr_vec, Z, idx, box):
        h = jax.vmap(self.readout)(g)

        is_ensemble = False
        if jnp.size(h, axis=1) > 1:
            # ensemble detected
            is_ensemble = True
            n_ens = jnp.size(h, axis=1)
            h = h[..., None]
            h = jnp.transpose(h, (1, 0, 2))

        p_i = h * self.scale[Z] + self.shift_param[Z]

        if self.mode == "l0":
            p_i = p_i
        elif self.mode == "l1":
            Rc = R - jnp.mean(R, axis=0, keepdims=True)
            r_hat = Rc / jnp.linalg.norm(Rc, axis=1)[:, None]
            p_i = p_i * R
        elif self.mode == "symmetric_traceless_l2":
            Rc = R - jnp.mean(R, axis=0, keepdims=True)
            r_hat = Rc / jnp.linalg.norm(Rc, axis=1)[:, None]
            r_rt = jnp.einsum("ni, nj -> nij", r_hat, r_hat)
            I = jnp.eye(3)
            symmetrized = 3 * r_rt - I
            p_i = p_i[..., None] * symmetrized
        else:
            raise KeyError("unknown symmetry option")

        if is_ensemble:
            p_i = jnp.swapaxes(p_i, 0, 1)  # natoms, nens, features...

        if self.apply_mask:
            p_i = mask_by_atom(p_i, Z)

        if self.aggregation == "none":
            result = p_i
        elif self.aggregation == "sum":
            result = fp64_sum(p_i, axis=0)
        elif self.aggregation == "mean":
            natoms = R.shape[0]
            result = fp64_sum(p_i, axis=0) / natoms
        else:
            raise KeyError("unknown aggregation")

        output = {self.pname: result}

        if is_ensemble:
            divisor = 1 / (n_ens - 1)
            if self.aggregation == "none":
                result = jnp.swapaxes(result, 0, 1)

            mean = jnp.mean(result, axis=0)
            uncertainty = divisor * fp64_sum((mean - result) ** 2, axis=0)
            output[self.pname] = mean
            output[self.pname + "_uncertainty"] = uncertainty

        return output



import os
import urllib
import zipfile


def download_md22_stachyose(data_path):
    url = "http://www.quantum-machine.org/gdml/repo/static/md22_stachyose.zip"
    file_path = data_path / "md22_stachyose.zip"

    os.makedirs(data_path, exist_ok=True)
    urllib.request.urlretrieve(url, file_path)

    with zipfile.ZipFile(file_path, "r") as zip_ref:
        zip_ref.extractall(data_path)

    file_path = modify_xyz_file(
        file_path.with_suffix(".xyz"), target_string="Energy", replacement_string="energy"
    )

    return file_path


def download_benzene_DFT(data_path):
    url = "http://www.quantum-machine.org/gdml/data/xyz/benzene2018_dft.zip"
    file_path = data_path / "benzene2018_dft.zip"

    os.makedirs(data_path, exist_ok=True)
    urllib.request.urlretrieve(url, file_path)

    with zipfile.ZipFile(file_path, "r") as zip_ref:
        zip_ref.extractall(data_path)

    new_file_path = data_path / "benzene.xyz"
    os.remove(file_path)

    return new_file_path


def download_etoh_ccsdt(data_path):
    url = "http://www.quantum-machine.org/gdml/data/xyz/ethanol_ccsd_t.zip"
    file_path = data_path / "ethanol_ccsd_t.zip"

    os.makedirs(data_path, exist_ok=True)
    urllib.request.urlretrieve(url, file_path)

    with zipfile.ZipFile(file_path, "r") as zip_ref:
        zip_ref.extractall(data_path)

    test_file_path = data_path / "ethanol_ccsd_t-test.xyz"
    train_file_path = data_path / "ethanol_ccsd_t-train.xyz"
    os.remove(file_path)

    return train_file_path, test_file_path


def download_md22_benzene_CCSDT(data_path):
    url = "http://www.quantum-machine.org/gdml/data/xyz/benzene_ccsd_t.zip"
    file_path = data_path / "benzene_ccsdt.zip"

    os.makedirs(data_path, exist_ok=True)
    urllib.request.urlretrieve(url, file_path)

    with zipfile.ZipFile(file_path, "r") as zip_ref:
        zip_ref.extractall(data_path)

    train_file_path = data_path / "benzene_ccsd_t-train.xyz"
    test_file_path = data_path / "benzene_ccsd_t-test.xyz"
    os.remove(file_path)

    return train_file_path, test_file_path


def modify_xyz_file(file_path, target_string, replacement_string):
    new_file_path = file_path.with_name(file_path.stem + "_mod" + file_path.suffix)

    with open(file_path, "r") as input_file, open(new_file_path, "w") as output_file:
        for line in input_file:
            # Replace all occurrences of the target string with the replacement string
            modified_line = line.replace(target_string, replacement_string)
            output_file.write(modified_line)
    return new_file_path


def mod_md_datasets(file_path):
    new_file_path = file_path.with_name(file_path.stem + "_mod" + file_path.suffix)
    with open(file_path, "r") as input_file, open(new_file_path, "w") as output_file:
        for line in input_file:
            if line.startswith("-"):
                modified_line = f"Properties=species:S:1:pos:R:3:forces:R:3 energy={line}"
            else:
                modified_line = line
            output_file.write(modified_line)

    os.remove(file_path)

    return new_file_path



from functools import partial
from typing import List, Union

import jax
import numpy as np
from ase import Atoms
from click import Path
from flax.core.frozen_dict import FrozenDict
from tqdm import trange

from apax.bal import feature_maps, kernel, selection, transforms
from apax.data.input_pipeline import OTFInMemoryDataset
from apax.nn.models import EnergyModel
from apax.train.checkpoints import (
    canonicalize_energy_model_parameters,
    check_for_ensemble,
    restore_parameters,
)
from apax.utils.transform import make_energy_only_model


def create_feature_fn(
    model: EnergyModel,
    params: FrozenDict,
    base_feature_map: feature_maps.FeatureTransformation,
    feature_transforms=[],
    is_ensemble: bool = False,
):
    """
    Converts a model into a feature map and transforms it as needed and
    sets it up for use in copmuting the features of a dataset.

    All transformations are applied on the feature function, not on computed features.
    Only the final function is jit compiled.


    Attributes
    ----------
    model: EnergyModel
        Model to be transformed.
    params: FrozenDict
        Model parameters
    base_feature_map: FeatureTransformation
        Class that transforms the model into a `FeatureMap`
    is_ensemble: bool
        Whether or not to apply the ensemble transformation i.e.
        an averaging of kernels for model ensembles.
    """
    feature_fn = base_feature_map.apply(model)

    if is_ensemble:
        feature_fn = transforms.ensemble_features(feature_fn)

    for transform in feature_transforms:
        feature_fn = transform.apply(feature_fn)

    feature_fn = transforms.batch_features(feature_fn)
    feature_fn = partial(feature_fn, params)
    feature_fn = jax.jit(feature_fn)
    return feature_fn


def compute_features(
    feature_fn: feature_maps.FeatureMap, dataset: OTFInMemoryDataset
) -> np.ndarray:
    """Compute the features of a dataset.

    Attributes
    ----------
    feature_fn:
        Function to compute the features with.
    dataset:
        Dataset to compute the features for.
    """
    features = []
    n_data = dataset.n_data
    ds = dataset.batch()

    pbar = trange(n_data, desc="Computing features", ncols=100, leave=True)
    for inputs in ds:
        g = feature_fn(inputs)
        features.append(np.asarray(g))
        pbar.update(g.shape[0])
    pbar.close()

    features = np.concatenate(features, axis=0)
    return features


def kernel_selection(
    model_dir: Union[Path, List[Path]],
    train_atoms: List[Atoms],
    pool_atoms: List[Atoms],
    base_fm_options: dict,
    selection_method: str,
    feature_transforms: list = [],
    selection_batch_size: int = 10,
    processing_batch_size: int = 64,
) -> list[int]:
    """
    Main function to facilitate batch data selection.
    Currently only the last layer gradient features and MaxDist selection method
    are available.
    More can be added as needed as this function is agnostic of the feature
    map/selection method internals.

    Attributes
    ----------
    model_dir: Union[Path, List[Path]]
        Path to the trained model or models which should be used to compute features.
    train_atoms: List[Atoms]
        List of `ase.Atoms` used to train the models.
    pool_atoms: List[Atoms]
        List of `ase.Atoms` to select new data from.
    base_fm_options:
        Settings for the base feature map.
    selection_method:
        Currently only "max_dist" is supported.
    feature_transforms:
        Feature transforms to be applied on top of the
        base feature map transform.
        Examples would include multiplication with or addition of a constant.
    selection_batch_size:
        Amount of new data points to be selected from `pool_atoms`.
    processing_batch_size:
        Amount of data points to compute the features for at once.
        Does not effect results, just the speed of processing.
    """
    selection_fn = {
        "max_dist": selection.max_dist_selection,
    }[selection_method]

    base_feature_map = feature_maps.FeatureMapOptions(base_fm_options)

    config, params = restore_parameters(model_dir)
    params = canonicalize_energy_model_parameters(params)
    n_models = check_for_ensemble(params)
    is_ensemble = n_models > 1

    n_train = len(train_atoms)
    dataset = OTFInMemoryDataset(
        train_atoms + pool_atoms,
        cutoff=config.model.basis.r_max,
        bs=processing_batch_size,
        n_epochs=1,
        ignore_labels=True,
        pos_unit=config.data.pos_unit,
        energy_unit=config.data.energy_unit,
    )

    _, init_box = dataset.init_input()

    Builder = config.model.get_builder()
    builder = Builder(config.model.model_dump(), n_species=119)

    energy_model = builder.build_energy_model(apply_mask=True, init_box=init_box)
    energy_model = make_energy_only_model(energy_model.apply)

    feature_fn = create_feature_fn(
        energy_model, params, base_feature_map, feature_transforms, is_ensemble
    )
    g = compute_features(feature_fn, dataset)
    km = kernel.KernelMatrix(g, n_train)
    ranking_indices, distances = selection_fn(km, selection_batch_size)

    g_train = g[:n_train]
    g_pool = g[n_train:]

    return ranking_indices, distances, g_train, g_pool



import einops
import jax


def geometric_moments(radial_function, dn, idx_i, n_atoms=None):
    # dn shape: neighbors x 3
    # radial_function shape: n_neighbors x n_radial

    # s = spatial dim = 3
    xyz = einops.repeat(dn, "nbrs s -> nbrs 1 s")
    xyz2 = einops.repeat(dn, "nbrs s -> nbrs 1 1 s")
    xyz3 = einops.repeat(dn, "nbrs s -> nbrs 1 1 1 s")

    # shape: n_neighbors x n_radial x (3)^(moment_number)
    # s_i = spatial = 3
    zero_moment = radial_function
    first_moment = einops.repeat(zero_moment, "n r -> n r 1") * xyz
    second_moment = einops.repeat(first_moment, "n r s1 -> n r s1 1") * xyz2
    third_moment = einops.repeat(second_moment, "n r s1 s2 -> n r s1 s2 1") * xyz3

    # shape: n_atoms x n_radial x (3)^(moment_number)
    zero_moment = jax.ops.segment_sum(zero_moment, idx_i, n_atoms)
    first_moment = jax.ops.segment_sum(first_moment, idx_i, n_atoms)
    second_moment = jax.ops.segment_sum(second_moment, idx_i, n_atoms)
    third_moment = jax.ops.segment_sum(third_moment, idx_i, n_atoms)

    moments = [zero_moment, first_moment, second_moment, third_moment]

    return moments



import importlib.metadata
import importlib.resources as pkg_resources
import json
import sys
from pathlib import Path

import typer
import yaml
from pydantic import ValidationError
from rich.console import Console

from apax.cli import templates

console = Console(highlight=False)

app = typer.Typer(
    context_settings={"help_option_names": ["-h", "--help"]},
    pretty_exceptions_show_locals=False,
)
validate_app = typer.Typer(
    pretty_exceptions_show_locals=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Validate training or MD config files.",
)
template_app = typer.Typer(
    pretty_exceptions_show_locals=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Create configuration file templates.",
)
app.add_typer(validate_app, name="validate")
app.add_typer(template_app, name="template")


@app.command()
def train(
    train_config_path: Path = typer.Argument(
        ..., help="Training configuration YAML file."
    ),
    log_level: str = typer.Option("info", help="Sets the training logging level."),
):
    """
    Starts the training of a model with parameters provided by a configuration file.
    """
    from apax.train.run import run

    run(train_config_path, log_level)


@app.command()
def md(
    train_config_path: Path = typer.Argument(
        ..., help="Configuration YAML file that was used to train a model."
    ),
    md_config_path: Path = typer.Argument(..., help="MD configuration YAML file."),
    log_level: str = typer.Option("info", help="Sets the training logging level."),
):
    """
    Starts performing a molecular dynamics simulation (currently only NHC thermostat)
    with parameters provided by a configuration file.
    """
    from apax.md import run_md

    run_md(train_config_path, md_config_path, log_level)


@app.command()
def eval(
    train_config_path: Path = typer.Argument(
        ..., help="Configuration YAML file that was used to train a model."
    ),
    n_data: int = typer.Option(
        -1,
        help=(
            "Number of test structures. (All structures are selected by not specifying"
            " it) Gets ignored if test_data_path is specified"
        ),
    ),
):
    """
    Starts performing the evaluation of the test dataset
    with parameters provided by a configuration file.
    """
    from apax.train.eval import eval_model

    eval_model(train_config_path, n_data)


@app.command()
def docs():
    """
    Opens the documentation website in your browser.
    """
    console.print("Opening apax's docs at https://apax.readthedocs.io/en/latest/")
    typer.launch("https://apax.readthedocs.io/en/latest/")


@app.command()
def schema():
    """
    Generating JSON schemata for autocompletion of train/md inputs in VSCode.
    """
    console.print("Generating JSON schema")
    from apax.config import Config, MDConfig

    vscode_path = Path(".vscode")
    vscode_path.mkdir(exist_ok=True)

    settings_path = vscode_path / "settings.json"

    if not settings_path.is_file():
        with settings_path.open("w") as f:
            json.dump({"yaml.schemas": {}}, f, indent=2)

    with settings_path.open("r") as f:
        settings = json.load(f)

    if "yaml.schemas" not in settings.keys():
        settings["yaml.schemas"] = {}

    train_schema = (vscode_path / "apaxtrain.schema.json").resolve().as_posix()
    md_schema = (vscode_path / "apaxmd.schema.json").resolve().as_posix()

    schemas = {train_schema: ["train*.yaml"], md_schema: ["md*.yaml"]}
    settings["yaml.schemas"].update(schemas)

    with settings_path.open("w") as f:
        json.dump(settings, f, indent=2)

    train_schema = Config.model_json_schema()
    md_schema = MDConfig.model_json_schema()
    with (vscode_path / "apaxtrain.schema.json").open("w") as f:
        json.dump(train_schema, f, indent=2)

    with (vscode_path / "apaxmd.schema.json").open("w") as f:
        json.dump(md_schema, f, indent=2)


def format_error(error):
    input_type = type(error["input"]).__name__
    loc = ".".join(error["loc"])
    error_msg = error["msg"]
    msg = f"{loc}\n  {error_msg}\n  input_type: {input_type}"
    if type(error["input"]) not in [dict, list]:
        field_input = error["input"]
        msg += f"\n  input: {field_input}"
    msg += "\n"
    return msg


def cleanup_error(e):
    e_clean = []
    for error in e.errors():
        error.pop("url")
        e_clean.append(format_error(error))
    e_clean = "".join(e_clean)
    return e_clean


@validate_app.command("train")
def validate_train_config(
    config_path: Path = typer.Argument(
        ..., help="Configuration YAML file to be validated."
    ),
):
    """
    Validates a training configuration file.

    Parameters
    ----------
    config_path: Path to the training configuration file.
    """
    from apax.config import Config

    with open(config_path, "r") as stream:
        user_config = yaml.safe_load(stream)

    try:
        _ = Config.model_validate(user_config)
    except ValidationError as e:
        print(f"{e.error_count()} validation errors for config")
        print(cleanup_error(e))
        console.print("Configuration Invalid!", style="red3")
        raise typer.Exit(code=1)
    else:
        console.print("Success!", style="green3")
        console.print(f"{config_path} is a valid training config.")


@validate_app.command("md")
def validate_md_config(
    config_path: Path = typer.Argument(
        ..., help="Configuration YAML file to be validated."
    ),
):
    """
    Validates a molecular dynamics configuration file.

    Parameters
    ----------
    config_path: Path to the molecular dynamics configuration file.
    """
    from apax.config import MDConfig

    with open(config_path, "r") as stream:
        user_config = yaml.safe_load(stream)

    try:
        _ = MDConfig.model_validate(user_config)
    except ValidationError as e:
        print(e)
        console.print("Configuration Invalid!", style="red3")
        raise typer.Exit(code=1)
    else:
        console.print("Success!", style="green3")
        console.print(f"{config_path} is a valid MD config.")


@app.command("visualize")
def visualize_model(
    config_path: Path = typer.Argument(
        ...,
        help=(
            "Training configuration file to be visualized. A CO molecule is taken as"
            " sample input."
        ),
    ),
):
    """
    Visualize a model based on a configuration file.
    A CO molecule is taken as sample input (influences number of atoms,
    number of species is set to 10).

    Parameters
    ----------
    config_path: Path to the training configuration file.
    """
    import jax

    from apax.config import Config
    from apax.utils.data import make_minimal_input

    with open(config_path, "r") as stream:
        user_config = yaml.safe_load(stream)

    try:
        config = Config.model_validate(user_config)
    except ValidationError as e:
        print(e)
        console.print("Configuration Invalid!", style="red3")
        raise typer.Exit(code=1)

    R, Z, idx, box, offsets = make_minimal_input()
    Builder = config.model.get_builder()
    builder = Builder(config.model.model_dump(), n_species=10)
    model = builder.build_energy_model()
    print(model.tabulate(jax.random.PRNGKey(0), R, Z, idx, box, offsets))


@template_app.command("train")
def template_train_config(
    full: bool = typer.Option(False, help="Use all input options."),
):
    """
    Creates a training input template in the current working directory.
    """
    if full:
        template_file = "train_config_full.yaml"
        config_path = "config_full.yaml"
    else:
        template_file = "train_config_minimal.yaml"
        config_path = "config.yaml"

    template_content = pkg_resources.read_text(templates, template_file)

    if Path(config_path).is_file():
        console.print("There is already a config file in the working directory.")
        sys.exit(1)
    else:
        with open(config_path, "w") as config:
            config.write(template_content)


@template_app.command("md")
def template_md_config():
    """
    Creates a training input template in the current working directory.
    """

    template_file = "md_config_minimal.yaml"
    config_path = "md_config.yaml"

    template_content = pkg_resources.read_text(templates, template_file)

    if Path(config_path).is_file():
        console.print("There is already a config file in the working directory.")
        sys.exit(1)
    else:
        with open(config_path, "w") as config:
            config.write(template_content)


def version_callback(value: bool) -> None:
    """Get the installed apax version."""
    if value:
        console.print(f"apax {importlib.metadata.version('apax')}")
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        None, "--version", "-V", callback=version_callback, is_eager=True
    ),
):
    # Taken from https://github.com/zincware/dask4dvc/blob/main/dask4dvc/cli/main.py
    _ = version



from typing import Iterable, Optional, Union

import jax.numpy as jnp
from jax import Array


def fp64_sum(
    X: Array, axis: Optional[Union[Iterable[int], int]] = None, keepdims: bool = False
):
    dtyp = jnp.float64
    result = jnp.sum(X, axis=axis, dtype=dtyp, keepdims=keepdims)
    return result


def normed_dotp(F_0, F_pred):
    F_0_norm = jnp.linalg.norm(F_0, ord=2, axis=2, keepdims=True)
    F_p_norm = jnp.linalg.norm(F_pred, ord=2, axis=2, keepdims=True)

    F_0_n = jnp.where(F_0_norm > 1e-6, F_0 / F_0_norm, jnp.zeros_like(F_0))
    F_p_n = jnp.where(F_p_norm > 1e-6, F_pred / F_p_norm, jnp.zeros_like(F_pred))

    dotp = jnp.einsum("bai, bai -> ba", F_0_n, F_p_n)
    return dotp



import logging
import os
from collections.abc import MutableMapping
from typing import Union

import yaml

from apax.config.md_config import MDConfig
from apax.config.train_config import Config

log = logging.getLogger(__name__)


def parse_config(config: Union[str, os.PathLike, dict], mode: str = "train") -> Config:
    """Load the training configuration from file or a dictionary.

    Parameters
    ----------
        config : str | os.PathLike | dict
            Path to the config file or a dictionary
            containing the config.
        mode: str, default = train
            Defines if the config is validated for training ("train")
            or MD simulation("md").
    """
    if isinstance(config, (str, os.PathLike)):
        with open(config, "r") as stream:
            config = yaml.safe_load(stream)

    if mode == "train":
        config = Config.model_validate(config)
    elif mode == "md":
        config = MDConfig.model_validate(config)

    return config


def flatten(dictionary, parent_key="", separator="_"):
    """https://stackoverflow.com/questions/6027558/
    flatten-nested-dictionaries-compressing-keys
    """
    items = []
    for key, value in dictionary.items():
        new_key = parent_key + separator + key if parent_key else key
        if isinstance(value, MutableMapping):
            items.extend(flatten(value, new_key, separator=separator).items())
        else:
            items.append((new_key, value))
    return dict(items)



import jax
import jax.numpy as jnp
import numpy as np

from apax.layers.masking import mask_by_atom, mask_by_neighbor


def make_atomic_prediction(n_atoms, n_total):
    n_padding = n_total - n_atoms

    Z = jnp.concatenate([np.ones(n_atoms), np.zeros(n_padding)])
    atomic_prediction = jnp.ones((n_total, 1))

    return atomic_prediction, Z


def make_neighbor_prediction(n_nbrs, n_total):
    n_padding = n_total - n_nbrs
    radial_func = np.ones((n_nbrs, 3))
    radial_func_padded = np.concatenate([radial_func, np.full((n_padding, 3), 20)])

    idx_i = np.arange(start=0, stop=n_nbrs)
    idx_j = np.arange(start=0, stop=n_nbrs)[::-1]
    idx = np.stack([idx_i, idx_j])

    idx_padding = np.zeros((2, n_padding))
    idx_padded = np.concatenate([idx, idx_padding], axis=1)

    return radial_func, radial_func_padded, idx_padded


def test_mask_by_atom():
    n_atoms = np.array([4, 6, 10])
    n_total = 10

    Zs = []
    preds = []
    for n in n_atoms:
        prediction, Z = make_atomic_prediction(n, n_total)
        preds.append(prediction)
        Zs.append(Z)

    preds = jnp.stack(preds, axis=0)
    Zs = jnp.stack(Zs, axis=0)
    assert np.all((np.sum(preds, axis=2) - n_total) < 1e-6)

    batched_mask_fn = jax.vmap(mask_by_atom, 0, 0)
    masked_preds = batched_mask_fn(preds[0][None, ...], Zs[0][None, ...])

    assert np.all((np.sum(masked_preds, axis=(1, 2)) - n_atoms) < 1e-6)


def test_mask_by_neighbor():
    n_nbrs = np.array([4, 6, 10])
    n_total = 10

    rfs = []
    rfs_padded = []
    idxs = []
    for n in n_nbrs:
        radial_func, radial_func_padded, idx_padded = make_neighbor_prediction(n, n_total)
        rfs.append(radial_func)
        rfs_padded.append(radial_func_padded)
        idxs.append(idx_padded)

    rfs_padded = jnp.stack(rfs_padded)
    idxs = jnp.stack(idxs)

    for rf, n in zip(rfs, n_nbrs):
        assert (jnp.sum(rf) - n * rf.shape[1]) < 1e-6

    batched_mask_fn = jax.vmap(mask_by_neighbor)

    masked_rfs = batched_mask_fn(rfs_padded, idxs)

    assert np.all(
        (jnp.sum(masked_rfs, axis=(1, 2)) - n_nbrs * masked_rfs.shape[2]) < 1e-6
    )



from typing import Any, Union

import einops
import flax.linen as nn
import jax.numpy as jnp

from apax.utils.convert import str_to_dtype


class PerElementScaleShift(nn.Module):
    n_species: int = 119
    scale: Union[jnp.array, float] = 1.0
    shift: Union[jnp.array, float] = 0.0
    dtype: Any = jnp.float32

    def setup(self):
        scale = jnp.asarray(self.scale)
        shift = jnp.asarray(self.shift)

        if len(scale.shape) > 0:
            n_species = scale.shape[0]
        else:
            n_species = self.n_species

        if len(scale.shape) == 1:
            scale = einops.repeat(scale, "species -> species 1")
        if len(shift.shape) == 1:
            shift = einops.repeat(shift, "species -> species 1")
        scale_init = nn.initializers.constant(scale)
        shift_init = nn.initializers.constant(shift)

        dtype = str_to_dtype(self.dtype)
        self.scale_param = self.param(
            "scale_per_element", scale_init, (n_species, 1), dtype
        )
        self.shift_param = self.param(
            "shift_per_element", shift_init, (n_species, 1), dtype
        )

    def __call__(self, x, Z):
        dtype = str_to_dtype(self.dtype)
        # x shape: n_atoms x 1
        # Z shape: n_atoms
        # scale[Z] shape: n_atoms x 1
        x = x.astype(dtype)

        out = self.scale_param[Z] * x + self.shift_param[Z]

        assert out.dtype == dtype
        return out



import logging
import pathlib

import ase
import h5py
import znh5md
import zntrack

from apax.nodes.model import Apax

log = logging.getLogger(__name__)


class ApaxBatchPrediction(zntrack.Node):
    """Create and Save the predictions from model on atoms.

    Attributes
    ----------
    model: Apax
        The Apax model node that implements the 'predict' method
    data: list[Atoms
        Atoms to predict properties for
    batch_size: int
        Number of structures to process at once.
        Descriptor fp64 recommended for large batch sizes

    predictions: list[Atoms] the atoms that have the predicted properties from model
    """

    data: list[ase.Atoms] = zntrack.deps()

    model: Apax = zntrack.deps()
    batch_size: int = zntrack.params(1)
    frames_path: pathlib.Path = zntrack.outs_path(zntrack.nwd / "frames.h5")

    def run(self):
        calc = self.model.get_calculator()
        frames = calc.batch_eval(self.data, self.batch_size)
        znh5md.write(self.frames_path, frames)

    @property
    def frames(self) -> list[ase.Atoms]:
        with self.state.fs.open(self.frames_path, "rb") as f:
            with h5py.File(f, "r") as h5:
                return znh5md.IO(file_handle=h5)[:]



import collections
import itertools
import logging

import jax
import jax.numpy as jnp
import numpy as np
from jax import tree_util
from vesin import NeighborList

log = logging.getLogger(__name__)


def compute_nl(positions, box, r_max):
    """
    Computes the neighbor list for a single structure.
    For periodic systems, positions are assumed to be in
    fractional coordinates.

    Parameters
    ----------
    positions : np.ndarray
        Positions of atoms.
    box : np.ndarray
        Simulation box dimensions.
    r_max : float
        Maximum interaction radius.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple containing neighbor indices array and offsets array.

    """
    if np.all(box < 1e-6):
        box, _ = get_shrink_wrapped_cell(positions)
        calculator = NeighborList(cutoff=r_max, full_list=True)
        idxs_i, idxs_j = calculator.compute(
            points=positions, box=box, periodic=False, quantities="ij"
        )

        neighbor_idxs = np.array([idxs_i, idxs_j], dtype=np.int16)

        n_neighbors = neighbor_idxs.shape[1]
        offsets = np.full([n_neighbors, 3], 0)

    else:
        positions = positions @ box.T
        calculator = NeighborList(cutoff=r_max, full_list=True)
        idxs_i, idxs_j, offsets = calculator.compute(
            points=positions, box=box.T, periodic=True, quantities="ijS"
        )
        neighbor_idxs = np.array([idxs_i, idxs_j], dtype=np.int32)
        offsets = np.matmul(offsets, box.T)
    return neighbor_idxs, offsets


def get_shrink_wrapped_cell(positions):
    """
    Get the shrink-wrapped simulation cell based on atomic positions.

    Parameters
    ----------
    positions : np.ndarray
        Atomic positions.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Tuple containing the shrink-wrapped cell matrix and origin.
    """
    rmin = np.min(positions, axis=0)
    rmax = np.max(positions, axis=0)
    cell_origin = rmin
    cell = np.diag(rmax - rmin)
    for idx in range(3):
        if cell[idx, idx] < 10e-1:
            cell[idx, idx] = 1.0

    cell[np.diag_indices_from(cell)] += 1

    return cell, cell_origin


def prefetch_to_single_device(iterator, size: int, sharding=None, n_step_jit=False):
    """
    inspired by
    https://flax.readthedocs.io/en/latest/_modules/flax/jax_utils.html#prefetch_to_device
    except it does not shard the data.
    """
    queue = collections.deque()

    if sharding:
        n_devices = len(sharding._devices)
        slice_start = 1
        shape = [n_devices]
        if n_step_jit:
            # replicate over multi-batch axis
            # data shape: njit x bs x ...
            slice_start = 2
            shape.insert(0, 1)

    def _prefetch(x: jax.Array):
        if sharding:
            remaining_axes = [1] * len(x.shape[slice_start:])
            final_shape = tuple(shape + remaining_axes)
            x = jax.device_put(x, sharding.reshape(final_shape))
        else:
            x = jnp.asarray(x)
        return x

    def enqueue(n):
        for data in itertools.islice(iterator, n):
            queue.append(tree_util.tree_map(_prefetch, data))

    enqueue(size)
    while queue:
        yield queue.popleft()
        enqueue(1)



import jax
import jax.numpy as jnp
import numpy as np

from apax.layers.descriptor.gaussian_moment_descriptor import GaussianMomentDescriptor


def test_gaussian_moment_descriptor():
    dR = np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [-1.0, 0.0, 0.0],
            [-1.0, 1.0, 0.0],
            [0.0, -1.0, 0.0],
            [1.0, -1.0, 0.0],
        ],
        dtype=jnp.float32,
    )
    Z = jnp.array([8, 1, 1])
    neighbor = jnp.array([[1, 2, 0, 2, 0, 1], [0, 0, 1, 1, 2, 2]])

    descriptor = GaussianMomentDescriptor()

    key = jax.random.PRNGKey(0)
    params = descriptor.init(key, dR, Z, neighbor)
    result = descriptor.apply(params, dR, Z, neighbor)
    result_jit = jax.jit(descriptor.apply)(params, dR, Z, neighbor)

    assert result.shape == (3, 360)
    assert result_jit.shape == (3, 360)



import logging
import os
from pathlib import Path
from typing import List, Literal, Optional, Union

import yaml
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    NonNegativeFloat,
    PositiveInt,
    create_model,
    model_validator,
)
from typing_extensions import Annotated

from apax.config.lr_config import CyclicCosineLR, LinearLR
from apax.config.model_config import GMNNConfig, ModelConfig
from apax.data.statistics import scale_method_list, shift_method_list

log = logging.getLogger(__name__)


class DatasetConfig(BaseModel, extra="forbid"):
    processing: str


class CachedDataset(DatasetConfig, extra="forbid"):
    """Dataset which pads everything (atoms, neighbors)
    to the largest system in the dataset.
    The NL is computed on the fly during the first epoch and stored to disk using
    tf.data's cache.
    Most performant option for datasets with samples of very similar size.

    Parameters
    ----------
    shuffle_buffer_size : int
        | Size of the buffer that is shuffled by tf.data.
        | Larger values require more RAM.
    """

    processing: Literal["cached"] = "cached"
    shuffle_buffer_size: PositiveInt = 1000


class OTFDataset(DatasetConfig, extra="forbid"):
    """Dataset which pads everything (atoms, neighbors)
    to the largest system in the dataset.
    The NL is computed on the fly and fed into a tf.data generator.
    Mostly for internal purposes.

    Parameters
    ----------
    shuffle_buffer_size : int
        | Size of the buffer that is shuffled by tf.data.
        | Larger values require more RAM.
    """

    processing: Literal["otf"] = "otf"
    shuffle_buffer_size: PositiveInt = 1000


class PBPDatset(DatasetConfig, extra="forbid"):
    """Dataset which pads everything (atoms, neighbors)
    to the next larges power of two.
    This limits the compute wasted due to padding at the (negligible)
    cost of some recompilations.
    The NL is computed on-the-fly in parallel for `num_workers` of batches.
    Does not use tf.data.

    Most performant option for datasets with significantly differently sized systems
    (e.g. MP, SPICE).

    Parameters
    ----------
    num_workers : int
        | Number of batches to be processed in parallel.
    atom_padding : int
        | Next nearest integer to which to pad per-atom arrays (positions, forces, ...).
    nl_padding: int
        | Next nearest integer to which to pad neighborlists.
    """

    processing: Literal["pbp"] = "pbp"
    num_workers: PositiveInt = 10
    atom_padding: PositiveInt = 10
    nl_padding: PositiveInt = 2000


class DataConfig(BaseModel, extra="forbid"):
    """
    Configuration for data loading, preprocessing and training.

    Parameters
    ----------
    directory : str, required
        | Path to directory where training results and checkpoints are written.
    experiment : str, required
        | Model name distinguishing from others in directory.
    data_path : str, required if train_data_path and val_data_path is not specified
        | Path to single dataset file.
    train_data_path : str, required if data_path is not specified
        | Path to training dataset.
    val_data_path : str, required if data_path is not specified
        | Path to validation dataset.
    test_data_path : str, optional
        | Path to test dataset.
    n_train : int, default = 1000
        | Number of training datapoints from `data_path`.
    n_valid : int, default = 100
        | Number of validation datapoints from `data_path`.
    batch_size : int, default = 32
        | Number of training examples to be evaluated at once.
    valid_batch_size : int, default = 100
        | Number of validation examples to be evaluated at once.
    shuffle_buffer_size : int, default = 1000
        | Size of the `tf.data` shuffle buffer.
    energy_regularisation :
        | Magnitude of the regularization in the per-element energy regression.
    pos_unit : str, default = "Ang"
        unit of length
    energy_unit : str, default = "eV"
        unit of energy
    """

    directory: str
    experiment: str
    dataset: Union[CachedDataset, OTFDataset, PBPDatset] = Field(
        CachedDataset(processing="cached"), discriminator="processing"
    )

    data_path: Optional[str] = None
    train_data_path: Optional[str] = None
    val_data_path: Optional[str] = None
    test_data_path: Optional[str] = None

    n_train: PositiveInt = 1000
    n_valid: PositiveInt = 100
    batch_size: PositiveInt = 32
    valid_batch_size: PositiveInt = 100

    shift_method: str = "per_element_regression_shift"
    shift_options: dict = {"energy_regularisation": 1.0}

    scale_method: str = "per_element_force_rms_scale"
    scale_options: Optional[dict] = {}

    pos_unit: Optional[str] = "Ang"
    energy_unit: Optional[str] = "eV"

    @model_validator(mode="after")
    def set_data_or_train_val_path(self):
        not_data_path = self.data_path is None
        not_train_path = self.train_data_path is None

        neither_set = not_data_path and not_train_path
        both_set = not not_data_path and not not_train_path

        if neither_set or both_set:
            raise ValueError("Please specify either data_path or train_data_path")

        return self

    @model_validator(mode="after")
    def validate_shift_scale_methods(self):
        method_lists = [shift_method_list, scale_method_list]
        requested_methods = [self.shift_method, self.scale_method]
        requested_options = [self.shift_options, self.scale_options]

        cases = zip(method_lists, requested_methods, requested_options)
        for method_list, requested_method, requested_params in cases:
            methods = {method.name: method for method in method_list}

            # check if method exists
            if requested_method not in methods.keys():
                raise KeyError(
                    f"The initialization method '{requested_method}' is not among the"
                    f" implemented methods. Choose from {methods.keys()}"
                )

            # check if parameters names are complete and correct
            method = methods[requested_method]
            fields = {
                name: (dtype, ...)
                for name, dtype in zip(method.parameters, method.dtypes)
            }
            MethodConfig = create_model(
                f"{method.name}Config", __config__=ConfigDict(extra="forbid"), **fields
            )

            _ = MethodConfig(**requested_params)

        return self

    @property
    def model_version_path(self):
        version_path = Path(self.directory) / self.experiment
        return version_path

    @property
    def best_model_path(self):
        return self.model_version_path / "best"


class OptimizerConfig(BaseModel, frozen=True, extra="forbid"):
    """
    Configuration of the optimizer.
    Learning rates of 0 will freeze the respective parameters.

    Parameters
    ----------
    name : str, default = "adam"
        Name of the optimizer. Can be any `optax` optimizer.
    emb_lr : NonNegativeFloat, default = 0.001
        Learning rate of the elemental embedding contraction coefficients.
    nn_lr : NonNegativeFloat, default = 0.001
        Learning rate of the neural network parameters.
    scale_lr : NonNegativeFloat, default = 0.0001
        Learning rate of the elemental output scaling factors.
    shift_lr : NonNegativeFloat, default = 0.003
        Learning rate of the elemental output shifts.
    zbl_lr : NonNegativeFloat, default = 0.0001
        Learning rate of the ZBL correction parameters.
    rep_scale_lr : NonNegativeFloat, default = 0.001
        LR for the length scale of these exponential repulsion potential.
    rep_prefactor_lr : NonNegativeFloat, default = 0.0001
        LR for the strength of the exponential repulsion potential.
    gradient_clipping: NonNegativeFloat, default = 1000.0
        Per element Gradient clipping value.
        Default is so high that it effectively disabled.
    schedule : LRSchedule = LinearLR
        Learning rate schedule.
    kwargs : dict, default = {}
        Optimizer keyword arguments. Passed to the `optax` optimizer.
    """

    name: str = "adam"
    emb_lr: NonNegativeFloat = 0.001
    nn_lr: NonNegativeFloat = 0.001
    scale_lr: NonNegativeFloat = 0.0001
    shift_lr: NonNegativeFloat = 0.003
    zbl_lr: NonNegativeFloat = 0.0001
    rep_scale_lr: NonNegativeFloat = 0.001
    rep_prefactor_lr: NonNegativeFloat = 0.0001

    gradient_clipping: NonNegativeFloat = 1000.0

    schedule: Union[LinearLR, CyclicCosineLR] = Field(
        LinearLR(name="linear"), discriminator="name"
    )
    kwargs: dict = {}


class MetricsConfig(BaseModel, extra="forbid"):
    """
    Configuration for the metrics collected during training.

    Parameters
    ----------
    name : str
        Keyword of the quantity, e.g., 'energy'.
    reductions : List[str]
        List of reductions performed on the difference between target and predictions.
        Can be 'mae', 'mse', 'rmse' for energies and forces.
        For forces, 'angle' can also be used.
    """

    name: str
    reductions: List[str]


class LossConfig(BaseModel, extra="forbid"):
    """
    Configuration of the loss functions used during training.

    Parameters
    ----------
    name : str
        Keyword of the quantity, e.g., 'energy'.
    loss_type : str, optional
        Weighting scheme for atomic contributions. See the MLIP package
        for reference 10.1088/2632-2153/abc9fe for details, by default "mse".
    weight : NonNegativeFloat, optional
        Weighting factor in the overall loss function, by default 1.0.
    atoms_exponent : NonNegativeFloat, optional
        Exponent for atomic contributions weighting, by default 1.
    parameters : dict, optional
        Additional parameters for configuring the loss function, by default {}.

    Notes
    -----
    This class specifies the configuration of the loss functions used during training.
    """

    name: str
    loss_type: str = "mse"
    weight: NonNegativeFloat = 1.0
    atoms_exponent: NonNegativeFloat = 1
    parameters: dict = {}


class CSVCallback(BaseModel, frozen=True, extra="forbid"):
    """
    Configuration of the CSVCallback.

    Parameters
    ----------
    name: Keyword of the callback used..
    """

    name: Literal["csv"]


class TBCallback(BaseModel, frozen=True, extra="forbid"):
    """
    Configuration of the TensorBoard callback.

    Parameters
    ----------
    name: Keyword of the callback used..
    """

    name: Literal["tensorboard"]


class MLFlowCallback(BaseModel, frozen=True, extra="forbid"):
    """
    Configuration of the MLFlow callback.

    Parameters
    ----------
    name: Keyword of the callback used.
    experiment: Path to the MLFlow experiment, e.g. /Users/<user>/<my_experiment>
    """

    name: Literal["mlflow"]
    experiment: str


CallBack = Annotated[
    Union[CSVCallback, TBCallback, MLFlowCallback], Field(discriminator="name")
]


class TrainProgressbarConfig(BaseModel, extra="forbid"):
    """
    Configuration of progressbars.

    Parameters
    ----------
    disable_epoch_pbar: Set to True to disable the epoch progress bar.
    disable_batch_pbar: Set to True to disable the batch progress bar.
    """

    disable_epoch_pbar: bool = False
    disable_batch_pbar: bool = True


class CheckpointConfig(BaseModel, extra="forbid"):
    """
    Checkpoint configuration.

    Parameters
    ----------
    ckpt_interval: Number of epochs between checkpoints.
    base_model_checkpoint: Path to the folder containing a pre-trained model ckpt.
    reset_layers: List of layer names for which the parameters will be reinitialized.
    """

    ckpt_interval: PositiveInt = 500
    base_model_checkpoint: Optional[str] = None
    reset_layers: List[str] = []


class WeightAverage(BaseModel, extra="forbid"):
    """Applies an exponential moving average to model parameters.

    Parameters
    ----------
    ema_start : int, default = 1
        Epoch at which to start averaging models.
    alpha : float, default = 0.9
        How much of the new model to use. 1.0 would mean no averaging, 0.0 no updates.
    """

    ema_start: int = 0
    alpha: float = 0.9


class Config(BaseModel, frozen=True, extra="forbid"):
    """
    Main configuration of a apax training run. Parameter that are config classes will
    be generated by parsing the config.yaml file and are specified
    as shown :ref:`here <train_config>`:

    Example
    -------
    .. code-block:: yaml

        data:
            directory: models/
            experiment: apax
                .
                .

    Parameters
    ----------
    n_epochs : int, required
        | Number of training epochs.
    patience : int, optional
        | Number of epochs without improvement before trainings gets terminated.
    seed : int, default = 1
        | Random seed.
    n_jitted_steps : int, default = 1
        | Number of train batches to be processed in a compiled loop.
        | Can yield significant speedups for small structures or small batch sizes.
    data : :class:`.DataConfig`
        | Data configuration.
    model : :class:`.ModelConfig`
        | Model configuration.
    metrics : List of :class:`.MetricsConfig`
        | Metrics configuration.
    loss : List of :class:`.LossConfig`
        | Loss configuration.
    optimizer : :class:`.OptimizerConfig`
        | Loss optimizer configuration.
    weight_average : :class:`.WeightAverage`, optional
        | Options for averaging weights between epochs.
    callbacks : List of various CallBack classes
        | Possible callbacks are :class:`.CSVCallback`,
        | :class:`.TBCallback`, :class:`.MLFlowCallback`
    progress_bar : :class:`.TrainProgressbarConfig`
        | Progressbar configuration.
    checkpoints : :class:`.CheckpointConfig`
        | Checkpoint configuration.
    data_parallel : bool, default = True
        | Automatically uses all available GPUs for data parallel training.
        | Set to false to force single device training.
    """

    n_epochs: PositiveInt
    patience: Optional[PositiveInt] = None
    seed: int = 1
    n_jitted_steps: int = 1
    data_parallel: bool = True

    data: DataConfig
    model: ModelConfig = Field(GMNNConfig(name="gmnn"), discriminator="name")
    metrics: List[MetricsConfig] = []
    loss: List[LossConfig]
    optimizer: OptimizerConfig = OptimizerConfig()
    weight_average: Optional[WeightAverage] = None
    callbacks: List[CallBack] = [CSVCallback(name="csv")]
    progress_bar: TrainProgressbarConfig = TrainProgressbarConfig()
    checkpoints: CheckpointConfig = CheckpointConfig()

    def dump_config(self, save_path):
        """
        Writes the current config file to the specified directory.

        Parameters
        ----------
        save_path: Path to the directory.
        """
        with open(os.path.join(save_path, "config.yaml"), "w") as conf:
            yaml.dump(self.model_dump(), conf, default_flow_style=False)
